import copy
import sys
import os
import warnings
# For redirecting stdout.
if (sys.version_info > (3, 0)):
    from io import StringIO
else:
    from cStringIO import StringIO

import unittest
import nose
import numpy as np
from numpy import testing, pi

import yeadon.inertia as inertia
import yeadon.human as hum

warnings.filterwarnings('ignore', category=DeprecationWarning)

class TestHuman(unittest.TestCase):
    """Tests the :py:class:`Human` class."""

    male1meas = os.path.join(os.path.split(__file__)[0], '..', '..',
            'misc', 'samplemeasurements', 'male1.txt')

    def runTest(self):
        # NOTE : This allows one to run this test at the interpreter.
        pass

    def test_init_default_cfg(self):
        """Uses misc/samplemeasurements/male1.txt."""

        h = hum.Human(self.male1meas)
        meas = h.meas

        # Regression test.
        testing.assert_almost_equal(h.mass, 58.200488588422544)
        inertiaDes = np.zeros((3, 3))
        inertiaDes[0, 0] = 9.63093850
        inertiaDes[1, 1] = 9.99497872
        inertiaDes[2, 2] = 5.45117742e-01
        testing.assert_allclose(h.inertia, inertiaDes, atol=1e-15)
        testing.assert_allclose(h.center_of_mass,
                np.array([[0], [0], [1.19967938e-02]]), atol=1e-15)

        assert h.is_symmetric == True
        assert h.meas_mass == -1
        # reading measurements.
        assert len(h.meas) == 95

        # averaging limbs.
        testing.assert_almost_equal(meas['La2L'], meas['Lb2L'])
        testing.assert_almost_equal(meas['La3L'], meas['Lb3L'])
        testing.assert_almost_equal(meas['La4L'], meas['Lb4L'])
        testing.assert_almost_equal(meas['La5L'], meas['Lb5L'])
        testing.assert_almost_equal(meas['La6L'], meas['Lb6L'])
        testing.assert_almost_equal(meas['La7L'], meas['Lb7L'])
        testing.assert_almost_equal(meas['La0p'], meas['Lb0p'])
        testing.assert_almost_equal(meas['La1p'], meas['Lb1p'])
        testing.assert_almost_equal(meas['La2p'], meas['Lb2p'])
        testing.assert_almost_equal(meas['La3p'], meas['Lb3p'])
        testing.assert_almost_equal(meas['La4p'], meas['Lb4p'])
        testing.assert_almost_equal(meas['La5p'], meas['Lb5p'])
        testing.assert_almost_equal(meas['La6p'], meas['Lb6p'])
        testing.assert_almost_equal(meas['La7p'], meas['Lb7p'])
        testing.assert_almost_equal(meas['La4w'], meas['Lb4w'])
        testing.assert_almost_equal(meas['La5w'], meas['Lb5w'])
        testing.assert_almost_equal(meas['La6w'], meas['Lb6w'])
        testing.assert_almost_equal(meas['La7w'], meas['Lb7w'])

        testing.assert_almost_equal(meas['Lj1L'], meas['Lk1L'])
        testing.assert_almost_equal(meas['Lj3L'], meas['Lk3L'])
        testing.assert_almost_equal(meas['Lj4L'], meas['Lk4L'])
        testing.assert_almost_equal(meas['Lj5L'], meas['Lk5L'])
        testing.assert_almost_equal(meas['Lj6L'], meas['Lk6L'])
        testing.assert_almost_equal(meas['Lj8L'], meas['Lk8L'])
        testing.assert_almost_equal(meas['Lj9L'], meas['Lk9L'])
        testing.assert_almost_equal(meas['Lj1p'], meas['Lk1p'])
        testing.assert_almost_equal(meas['Lj2p'], meas['Lk2p'])
        testing.assert_almost_equal(meas['Lj3p'], meas['Lk3p'])
        testing.assert_almost_equal(meas['Lj4p'], meas['Lk4p'])
        testing.assert_almost_equal(meas['Lj5p'], meas['Lk5p'])
        testing.assert_almost_equal(meas['Lj6p'], meas['Lk6p'])
        testing.assert_almost_equal(meas['Lj7p'], meas['Lk7p'])
        testing.assert_almost_equal(meas['Lj8p'], meas['Lk8p'])
        testing.assert_almost_equal(meas['Lj9p'], meas['Lk9p'])
        testing.assert_almost_equal(meas['Lj8w'], meas['Lk8w'])
        testing.assert_almost_equal(meas['Lj9w'], meas['Lk9w'])
        testing.assert_almost_equal(meas['Lj6d'], meas['Lk6d'])

        # configuration.
        assert len(h.CFG) == 21
        CFGnames = ('somersault',
                    'tilt',
                    'twist',
                    'PTsagittalFlexion',
                    'PTbending',
                    'TCspinalTorsion',
                    'TCsagittalSpinalFlexion',
                    'CA1extension',
                    'CA1adduction',
                    'CA1rotation',
                    'CB1extension',
                    'CB1abduction',
                    'CB1rotation',
                    'A1A2extension',
                    'B1B2extension',
                    'PJ1extension',
                    'PJ1adduction',
                    'PK1extension',
                    'PK1abduction',
                    'J1J2flexion',
                    'K1K2flexion')
        for key in CFGnames:
            assert h.CFG[key] == 0.0

        # Check initialized global position and orientation.
        testing.assert_allclose(h._coord_sys_pos, np.array([[0, 0, 0]]).T)
        testing.assert_allclose(h._coord_sys_orient, np.eye(3))


        # Check that all segments exist.
        self.assertEquals(len(h.segments), 11)
        self.assertEquals(h.segments[0], h.P)
        self.assertEquals(h.segments[1], h.T)
        self.assertEquals(h.segments[2], h.C)
        self.assertEquals(h.segments[3], h.A1)
        self.assertEquals(h.segments[4], h.A2)
        self.assertEquals(h.segments[5], h.B1)
        self.assertEquals(h.segments[6], h.B2)
        self.assertEquals(h.segments[7], h.J1)
        self.assertEquals(h.segments[8], h.J2)
        self.assertEquals(h.segments[9], h.K1)
        self.assertEquals(h.segments[10], h.K2)

        # Ensure mass is unchanged from what it should be.
        testing.assert_almost_equal(h.mass,
                h.P.mass + h.T.mass + h.C.mass + h.A1.mass + h.A2.mass +
                h.B1.mass + h.B2.mass + h.J1.mass + h.J2.mass +
                h.K1.mass + h.K2.mass)

        # Initialize measurements using the dict from the previous one.
        h2 = hum.Human(h.meas)
        # Crude test...
        testing.assert_almost_equal(h.mass, h2.mass)

        # - Inspect symmetry.
        # Symmetric by default.
        self.assertTrue(h.is_symmetric)
        self.assertEqual(h.K1.mass, h.J1.mass)
        testing.assert_almost_equal(h.K2.mass, h.J2.mass)
        self.assertEqual(h.A1.mass, h.B1.mass)
        self.assertEqual(h.A2.mass, h.B2.mass)

    def test_density_set(self):
        """Tests the Chandler and Clauser density sets."""

        # Ensure the densities themselves have not regressed.
        segment_names = ['head-neck', 'shoulders', 'thorax', 'abdomen-pelvis',
                'upper-arm', 'forearm', 'hand', 'thigh', 'lower-leg', 'foot']
        segmental_densities_des = {
            'Chandler': dict(zip(segment_names,
            [1056,  853,  853,  853, 1005, 1052, 1080, 1020, 1078, 1091])),
            'Dempster': dict(zip(segment_names,
            [1110, 1040,  920, 1010, 1070, 1130, 1160, 1050, 1090, 1100])),
            'Clauser': dict(zip(segment_names,
            [1070, 1019, 1019, 1019, 1056, 1089, 1109, 1044, 1085, 1084])),
            }

        # Input error.
        self.assertRaises(Exception,
                hum.Human, self.male1meas, density_set='badname')
        try:
            hum.Human(self.male1meas, density_set='badname')
        except Exception as e:
            self.assertEquals(str(e), "Density set 'badname' is not one "
                    "of 'Chandler', 'Clauser', or 'Dempster'.")

        h = hum.Human(self.male1meas, density_set='Chandler')

        # Ensure the densities themselves have not regressed.
        segmental_densities_des = {
            'Chandler': dict(zip(segment_names,
            [1056,  853,  853,  853, 1005, 1052, 1080, 1020, 1078, 1091])),
            'Dempster': dict(zip(segment_names,
            [1110, 1040,  920, 1010, 1070, 1130, 1160, 1050, 1090, 1100])),
            'Clauser': dict(zip(segment_names,
            [1070, 1019, 1019, 1019, 1056, 1089, 1109, 1044, 1085, 1084])),
            }
        segmental_densities = h.segmental_densities
        for key, val in segmental_densities.items():
            for seg, dens in val.items():
                self.assertEquals(dens, segmental_densities_des[key][seg])

        # Regression test.
        testing.assert_almost_equal(h.mass, 54.639701113740323)
        inertiaDes = np.zeros((3, 3))
        inertiaDes[0, 0] = 9.26910316
        inertiaDes[1, 1] = 9.61346162
        inertiaDes[2, 2] = 0.512454528
        testing.assert_allclose(h.inertia, inertiaDes, atol=1e-15)
        testing.assert_allclose(h.center_of_mass,
                np.array([[0], [0], [2.08057425e-03]]), atol=1e-15)

        # Regression for the remaining density set.
        h = hum.Human(self.male1meas, density_set='Clauser')

        testing.assert_almost_equal(h.mass, 59.061501074879487)
        inertiaDes = np.zeros((3, 3))
        inertiaDes[0, 0] = 9.6382444
        inertiaDes[1, 1] = 10.002298
        inertiaDes[2, 2] = 0.550455953
        testing.assert_allclose(h.inertia, inertiaDes, atol=1e-15)
        testing.assert_allclose(h.center_of_mass,
                np.array([[0], [0], [0.0176605046]]), atol=1e-15)

    def test_init_symmetry_off(self):
        """Uses misc/samplemeasurements/male1.txt."""

        h = hum.Human(self.male1meas, symmetric=False)
        meas = h.meas

        self.assertFalse(h.is_symmetric)
        self.assertNotEqual(h.K1.mass, h.J1.mass)
        self.assertNotEqual(h.K2.mass, h.J2.mass)
        self.assertNotEqual(h.A1.mass, h.B1.mass)
        self.assertNotEqual(h.A2.mass, h.B2.mass)

        # Check level and solid definitions.
        # segment s
        self.assertEquals(len(h._Ls), 9)
        self.assertEquals(h._Ls[0].label, 'Ls0: hip joint centre')
        testing.assert_almost_equal(h._Ls[0].perimeter, meas['Ls0p'])
        testing.assert_almost_equal(h._Ls[0].width, meas['Ls0w'])

        self.assertEquals(h._Ls[1].label, 'Ls1: umbilicus')
        testing.assert_almost_equal(h._Ls[1].perimeter, meas['Ls1p'])
        testing.assert_almost_equal(h._Ls[1].width, meas['Ls1w'])

        self.assertEquals(h._Ls[2].label, 'Ls2: lowest front rib')
        testing.assert_almost_equal(h._Ls[2].perimeter, meas['Ls2p'])
        testing.assert_almost_equal(h._Ls[2].width, meas['Ls2w'])

        self.assertEquals(h._Ls[3].label, 'Ls3: nipple')
        testing.assert_almost_equal(h._Ls[3].perimeter, meas['Ls3p'])
        testing.assert_almost_equal(h._Ls[3].width, meas['Ls3w'])

        self.assertEquals(h._Ls[4].label, 'Ls4: shoulder joint centre')
        testing.assert_almost_equal(h._Ls[4].radius, meas['Ls4d'] / 2)
        testing.assert_almost_equal(h._Ls[4].width, meas['Ls4w'])

        # TODO Hard-coded parameters for the acromion from Yeadon's ISEG.
        self.assertEquals(h._Ls[5].label, 'Ls5: acromion')
        testing.assert_almost_equal(
                h._Ls[5].thickness, h._Ls[4].width / 2 - h._Ls[5].radius)
        testing.assert_almost_equal(h._Ls[5].radius, 0.57 * h._Ls[4].radius)

        self.assertEquals(h._Ls[6].label, 'Ls5: acromion/bottom of neck')
        testing.assert_almost_equal(h._Ls[6].perimeter, meas['Ls5p'])
        self.assertEquals(h._Ls[6].thickness, 0)

        self.assertEquals(h._Ls[7].label, 'Ls6: beneath nose')
        testing.assert_almost_equal(h._Ls[7].perimeter, meas['Ls6p'])
        self.assertEquals(h._Ls[7].thickness, 0)

        self.assertEquals(h._Ls[8].label, 'Ls7: above ear')
        testing.assert_almost_equal(h._Ls[8].perimeter, meas['Ls7p'])
        self.assertEquals(h._Ls[8].thickness, 0)

        self.assertEquals(len(h._s), 8)
        self.assertEquals(h._s[0].label, 's0: hip joint centre')
        testing.assert_almost_equal(h._s[0].density,
                h.segmental_densities['Dempster']['abdomen-pelvis'])
        self.assertEquals(h._s[1].label, 's1: umbilicus')
        testing.assert_almost_equal(h._s[1].density,
                h.segmental_densities['Dempster']['abdomen-pelvis'])
        self.assertEquals(h._s[2].label, 's2: lowest front rib')
        testing.assert_almost_equal(h._s[2].density,
                h.segmental_densities['Dempster']['thorax'])
        self.assertEquals(h._s[3].label, 's3: nipple')
        testing.assert_almost_equal(h._s[3].density,
                h.segmental_densities['Dempster']['thorax'])
        self.assertEquals(h._s[4].label, 's4: shoulder joint centre')
        testing.assert_almost_equal(h._s[4].density,
                h.segmental_densities['Dempster']['shoulders'])
        self.assertEquals(h._s[5].label, 's5: acromion')
        testing.assert_almost_equal(h._s[5].density,
                h.segmental_densities['Dempster']['head-neck'])
        self.assertEquals(h._s[6].label, 's6: beneath nose')
        testing.assert_almost_equal(h._s[6].density,
                h.segmental_densities['Dempster']['head-neck'])
        self.assertEquals(h._s[7].label, 's7: above ear')
        testing.assert_almost_equal(h._s[7].density,
                h.segmental_densities['Dempster']['head-neck'])

        # arms
        # 'a'
        self.assertEquals(len(h._La), 8)
        self.assertEquals(h._La[0].label, 'La0: shoulder joint centre')
        testing.assert_almost_equal(h._La[0].perimeter, meas['La0p'])
        self.assertEquals(h._La[0].thickness, 0)

        self.assertEquals(h._La[1].label, 'La1: mid-arm')
        testing.assert_almost_equal(h._La[1].perimeter, meas['La1p'])
        self.assertEquals(h._La[1].thickness, 0)

        self.assertEquals(h._La[2].label, 'La2: elbow joint centre')
        testing.assert_almost_equal(h._La[2].perimeter, meas['La2p'])
        self.assertEquals(h._La[2].thickness, 0)

        self.assertEquals(h._La[3].label, 'La3: maximum forearm perimeter')
        testing.assert_almost_equal(h._La[3].perimeter, meas['La3p'])
        self.assertEquals(h._La[3].thickness, 0)

        self.assertEquals(h._La[4].label, 'La4: wrist joint centre')
        testing.assert_almost_equal(h._La[4].perimeter, meas['La4p'])
        testing.assert_almost_equal(h._La[4].width, meas['La4w'])

        self.assertEquals(h._La[5].label, 'La5: base of thumb')
        testing.assert_almost_equal(h._La[5].perimeter, meas['La5p'])
        testing.assert_almost_equal(h._La[5].width, meas['La5w'])

        self.assertEquals(h._La[6].label, 'La6: knuckles')
        testing.assert_almost_equal(h._La[6].perimeter, meas['La6p'])
        testing.assert_almost_equal(h._La[6].width, meas['La6w'])

        self.assertEquals(h._La[7].label, 'La7: fingernails')
        testing.assert_almost_equal(h._La[7].perimeter, meas['La7p'])
        testing.assert_almost_equal(h._La[7].width, meas['La7w'])

        self.assertEquals(len(h._a_solids), 7)
        self.assertEquals(h._a_solids[0].label, 'a0: shoulder joint centre')
        testing.assert_almost_equal(h._a_solids[0].density,
                h.segmental_densities['Dempster']['upper-arm'])

        self.assertEquals(h._a_solids[1].label, 'a1: mid-arm')
        testing.assert_almost_equal(h._a_solids[1].density,
                h.segmental_densities['Dempster']['upper-arm'])

        self.assertEquals(h._a_solids[2].label, 'a2: elbow joint centre')
        testing.assert_almost_equal(h._a_solids[2].density,
                h.segmental_densities['Dempster']['forearm'])

        self.assertEquals(h._a_solids[3].label, 'a3: maximum forearm perimeter')
        testing.assert_almost_equal(h._a_solids[3].density,
                h.segmental_densities['Dempster']['forearm'])

        self.assertEquals(h._a_solids[4].label, 'a4: wrist joint centre')
        testing.assert_almost_equal(h._a_solids[4].density,
                h.segmental_densities['Dempster']['hand'])

        self.assertEquals(h._a_solids[5].label, 'a5: base of thumb')
        testing.assert_almost_equal(h._a_solids[5].density,
                h.segmental_densities['Dempster']['hand'])

        self.assertEquals(h._a_solids[6].label, 'a6: knuckles')
        testing.assert_almost_equal(h._a_solids[6].density,
                h.segmental_densities['Dempster']['hand'])

        # 'b'
        self.assertEquals(len(h._Lb), 8)
        self.assertEquals(h._Lb[0].label, 'Lb0: shoulder joint centre')
        testing.assert_almost_equal(h._Lb[0].perimeter, meas['Lb0p'])
        self.assertEquals(h._Lb[0].thickness, 0)

        self.assertEquals(h._Lb[1].label, 'Lb1: mid-arm')
        testing.assert_almost_equal(h._Lb[1].perimeter, meas['Lb1p'])
        self.assertEquals(h._Lb[1].thickness, 0)

        self.assertEquals(h._Lb[2].label, 'Lb2: elbow joint centre')
        testing.assert_almost_equal(h._Lb[2].perimeter, meas['Lb2p'])
        self.assertEquals(h._Lb[2].thickness, 0)

        self.assertEquals(h._Lb[3].label, 'Lb3: maximum forearm perimeter')
        testing.assert_almost_equal(h._Lb[3].perimeter, meas['Lb3p'])
        self.assertEquals(h._Lb[3].thickness, 0)

        self.assertEquals(h._Lb[4].label, 'Lb4: wrist joint centre')
        testing.assert_almost_equal(h._Lb[4].perimeter, meas['Lb4p'])
        testing.assert_almost_equal(h._Lb[4].width, meas['Lb4w'])

        self.assertEquals(h._Lb[5].label, 'Lb5: base of thumb')
        testing.assert_almost_equal(h._Lb[5].perimeter, meas['Lb5p'])
        testing.assert_almost_equal(h._Lb[5].width, meas['Lb5w'])

        self.assertEquals(h._Lb[6].label, 'Lb6: knuckles')
        testing.assert_almost_equal(h._Lb[6].perimeter, meas['Lb6p'])
        testing.assert_almost_equal(h._Lb[6].width, meas['Lb6w'])

        self.assertEquals(h._Lb[7].label, 'Lb7: fingernails')
        testing.assert_almost_equal(h._Lb[7].perimeter, meas['Lb7p'])
        testing.assert_almost_equal(h._Lb[7].width, meas['Lb7w'])

        self.assertEquals(len(h._b_solids), 7)
        self.assertEquals(h._b_solids[0].label, 'b0: shoulder joint centre')
        testing.assert_almost_equal(h._b_solids[0].density,
                h.segmental_densities['Dempster']['upper-arm'])

        self.assertEquals(h._b_solids[1].label, 'b1: mid-arm')
        testing.assert_almost_equal(h._b_solids[1].density,
                h.segmental_densities['Dempster']['upper-arm'])

        self.assertEquals(h._b_solids[2].label, 'b2: elbow joint centre')
        testing.assert_almost_equal(h._b_solids[2].density,
                h.segmental_densities['Dempster']['forearm'])

        self.assertEquals(h._b_solids[3].label, 'b3: maximum forearm perimeter')
        testing.assert_almost_equal(h._b_solids[3].density,
                h.segmental_densities['Dempster']['forearm'])

        self.assertEquals(h._b_solids[4].label, 'b4: wrist joint centre')
        testing.assert_almost_equal(h._b_solids[4].density,
                h.segmental_densities['Dempster']['hand'])

        self.assertEquals(h._b_solids[5].label, 'b5: base of thumb')
        testing.assert_almost_equal(h._b_solids[5].density,
                h.segmental_densities['Dempster']['hand'])

        self.assertEquals(h._b_solids[6].label, 'b6: knuckles')
        testing.assert_almost_equal(h._b_solids[6].density,
                h.segmental_densities['Dempster']['hand'])

        # legs
        # 'j'
        self.assertEquals(len(h._Lj), 10)
        Lj0p = np.pi * np.sqrt(h._Ls[0].radius * h._Ls[0].width)
        self.assertEquals(h._Lj[0].label, 'Lj0: hip joint centre')
        testing.assert_almost_equal(h._Lj[0].perimeter, Lj0p)
        self.assertEquals(h._Lj[0].thickness, 0)

        self.assertEquals(h._Lj[1].label, 'Lj1: crotch')
        testing.assert_almost_equal(h._Lj[1].perimeter, meas['Lj1p'])
        self.assertEquals(h._Lj[1].thickness, 0)

        self.assertEquals(h._Lj[2].label, 'Lj2: mid-thigh')
        testing.assert_almost_equal(h._Lj[2].perimeter, meas['Lj2p'])
        self.assertEquals(h._Lj[2].thickness, 0)

        self.assertEquals(h._Lj[3].label, 'Lj3: knee joint centre')
        testing.assert_almost_equal(h._Lj[3].perimeter, meas['Lj3p'])
        self.assertEquals(h._Lj[3].thickness, 0)

        self.assertEquals(h._Lj[4].label, 'Lj4: maximum calf perimeter')
        testing.assert_almost_equal(h._Lj[4].perimeter, meas['Lj4p'])
        self.assertEquals(h._Lj[4].thickness, 0)

        self.assertEquals(h._Lj[5].label, 'Lj5: ankle joint centre')
        testing.assert_almost_equal(h._Lj[5].perimeter, meas['Lj5p'])
        self.assertEquals(h._Lj[5].thickness, 0)

        self.assertEquals(h._Lj[6].label, 'Lj6: heel')
        testing.assert_almost_equal(h._Lj[6].perimeter, meas['Lj6p'])
        testing.assert_almost_equal(h._Lj[6].width, meas['Lj6d'])

        self.assertEquals(h._Lj[7].label, 'Lj7: arch')
        testing.assert_almost_equal(h._Lj[7].perimeter, meas['Lj7p'])

        self.assertEquals(h._Lj[8].label, 'Lj8: ball')
        testing.assert_almost_equal(h._Lj[8].perimeter, meas['Lj8p'])
        testing.assert_almost_equal(h._Lj[8].width, meas['Lj8w'])

        self.assertEquals(h._Lj[9].label, 'Lj9: toe nails')
        testing.assert_almost_equal(h._Lj[9].perimeter, meas['Lj9p'])
        testing.assert_almost_equal(h._Lj[9].width, meas['Lj9w'])

        self.assertEquals(len(h._j_solids), 9)
        self.assertEquals(h._j_solids[0].label, 'j0: hip joint centre')
        testing.assert_almost_equal(h._j_solids[0].density,
                h.segmental_densities['Dempster']['thigh'])

        self.assertEquals(h._j_solids[1].label, 'j1: crotch')
        testing.assert_almost_equal(h._j_solids[1].density,
                h.segmental_densities['Dempster']['thigh'])

        self.assertEquals(h._j_solids[2].label, 'j2: mid-thigh')
        testing.assert_almost_equal(h._j_solids[2].density,
                h.segmental_densities['Dempster']['thigh'])

        self.assertEquals(h._j_solids[3].label, 'j3: knee joint centre')
        testing.assert_almost_equal(h._j_solids[3].density,
                h.segmental_densities['Dempster']['lower-leg'])

        self.assertEquals(h._j_solids[4].label, 'j4: maximum calf perimeter')
        testing.assert_almost_equal(h._j_solids[4].density,
                h.segmental_densities['Dempster']['lower-leg'])

        self.assertEquals(h._j_solids[5].label, 'j5: ankle joint centre')
        testing.assert_almost_equal(h._j_solids[5].density,
                h.segmental_densities['Dempster']['foot'])

        self.assertEquals(h._j_solids[6].label, 'j6: heel')
        testing.assert_almost_equal(h._j_solids[6].density,
                h.segmental_densities['Dempster']['foot'])

        self.assertEquals(h._j_solids[7].label, 'j7: arch')
        testing.assert_almost_equal(h._j_solids[7].density,
                h.segmental_densities['Dempster']['foot'])

        self.assertEquals(h._j_solids[8].label, 'j8: ball')
        testing.assert_almost_equal(h._j_solids[8].density,
                h.segmental_densities['Dempster']['foot'])

        # 'k'
        self.assertEquals(len(h._Lk), 10)
        Lk0p = np.pi * np.sqrt(h._Ls[0].radius * h._Ls[0].width)
        self.assertEquals(h._Lk[0].label, 'Lk0: hip joint centre')
        testing.assert_almost_equal(h._Lk[0].perimeter, Lk0p)
        self.assertEquals(h._Lk[0].thickness, 0)

        self.assertEquals(h._Lk[1].label, 'Lk1: crotch')
        testing.assert_almost_equal(h._Lk[1].perimeter, meas['Lk1p'])
        self.assertEquals(h._Lk[1].thickness, 0)

        self.assertEquals(h._Lk[2].label, 'Lk2: mid-thigh')
        testing.assert_almost_equal(h._Lk[2].perimeter, meas['Lk2p'])
        self.assertEquals(h._Lk[2].thickness, 0)

        self.assertEquals(h._Lk[3].label, 'Lk3: knee joint centre')
        testing.assert_almost_equal(h._Lk[3].perimeter, meas['Lk3p'])
        self.assertEquals(h._Lk[3].thickness, 0)

        self.assertEquals(h._Lk[4].label, 'Lk4: maximum calf perimeter')
        testing.assert_almost_equal(h._Lk[4].perimeter, meas['Lk4p'])
        self.assertEquals(h._Lk[4].thickness, 0)

        self.assertEquals(h._Lk[5].label, 'Lk5: ankle joint centre')
        testing.assert_almost_equal(h._Lk[5].perimeter, meas['Lk5p'])
        self.assertEquals(h._Lk[5].thickness, 0)

        self.assertEquals(h._Lk[6].label, 'Lk6: heel')
        testing.assert_almost_equal(h._Lk[6].perimeter, meas['Lk6p'])
        testing.assert_almost_equal(h._Lk[6].width, meas['Lk6d'])

        self.assertEquals(h._Lk[7].label, 'Lk7: arch')
        testing.assert_almost_equal(h._Lk[7].perimeter, meas['Lk7p'])

        self.assertEquals(h._Lk[8].label, 'Lk8: ball')
        testing.assert_almost_equal(h._Lk[8].perimeter, meas['Lk8p'])
        testing.assert_almost_equal(h._Lk[8].width, meas['Lk8w'])

        self.assertEquals(h._Lk[9].label, 'Lk9: toe nails')
        testing.assert_almost_equal(h._Lk[9].perimeter, meas['Lk9p'])
        testing.assert_almost_equal(h._Lk[9].width, meas['Lk9w'])

        self.assertEquals(len(h._k_solids), 9)
        self.assertEquals(h._k_solids[0].label, 'k0: hip joint centre')
        testing.assert_almost_equal(h._k_solids[0].density,
                h.segmental_densities['Dempster']['thigh'])

        self.assertEquals(h._k_solids[1].label, 'k1: crotch')
        testing.assert_almost_equal(h._k_solids[1].density,
                h.segmental_densities['Dempster']['thigh'])

        self.assertEquals(h._k_solids[2].label, 'k2: mid-thigh')
        testing.assert_almost_equal(h._k_solids[2].density,
                h.segmental_densities['Dempster']['thigh'])

        self.assertEquals(h._k_solids[3].label, 'k3: knee joint centre')
        testing.assert_almost_equal(h._k_solids[3].density,
                h.segmental_densities['Dempster']['lower-leg'])

        self.assertEquals(h._k_solids[4].label, 'k4: maximum calf perimeter')
        testing.assert_almost_equal(h._k_solids[4].density,
                h.segmental_densities['Dempster']['lower-leg'])

        self.assertEquals(h._k_solids[5].label, 'k5: ankle joint centre')
        testing.assert_almost_equal(h._k_solids[5].density,
                h.segmental_densities['Dempster']['foot'])

        self.assertEquals(h._k_solids[6].label, 'k6: heel')
        testing.assert_almost_equal(h._k_solids[6].density,
                h.segmental_densities['Dempster']['foot'])

        self.assertEquals(h._k_solids[7].label, 'k7: arch')
        testing.assert_almost_equal(h._k_solids[7].density,
                h.segmental_densities['Dempster']['foot'])

        self.assertEquals(h._k_solids[8].label, 'k8: ball')
        testing.assert_almost_equal(h._k_solids[8].density,
                h.segmental_densities['Dempster']['foot'])

    def test_init_interesting_cfg(self):
        """Providing a dict for CFG, input errors, and out of bounds errors."""

        # Normal behavior.
        CFG = {'somersault': 0.0,
                'tilt': 0.0,
                'twist': np.pi/2,
                'PTsagittalFlexion': 0.0,
                'PTbending': 0.0,
                'TCspinalTorsion': 0.0,
                'TCsagittalSpinalFlexion': 0.0,
                'CA1extension': 0.0,
                'CA1adduction': 0.0,
                'CA1rotation': 0.0,
                'CB1extension': 0.0,
                'CB1abduction': np.pi/4,
                'CB1rotation': 0.0,
                'A1A2extension': 0.0,
                'B1B2extension': 0.0,
                'PJ1extension': 0.0,
                'PJ1adduction': 0.0,
                'PK1extension': 0.0,
                'PK1abduction': 0.0,
                'J1J2flexion': 0.0,
                'K1K2flexion': 0.0,
                }
        h = hum.Human(self.male1meas, CFG)
        self.assertEqual(h.CFG, CFG)

        # Missing a value.
        CFG = {'somersault': 0.0,
                'tilt': 0.0,
                'twist': np.pi/2,
                'PTsagittalFlexion': 0.0,
                'PTbending': 0.0,
                'TCspinalTorsion': 0.0,
                'TCsagittalSpinalFlexion': 0.0,
                'CA1extension': 0.0,
                'CA1adduction': 0.0,
                'CA1rotation': 0.0,
                'CB1extension': 0.0,
                'CB1abduction': np.pi/4,
                'CB1rotation': 0.0,
                'A1A2extension': 0.0,
                'PJ1extension': 0.0,
                'PJ1adduction': 0.0,
                'PK1extension': 0.0,
                'PK1abduction': 0.0,
                'J1J2flexion': 0.0,
                'K1K2flexion': 0.0,
                }
        self.assertRaises(Exception, hum.Human, self.male1meas, CFG)
        try:
            hum.Human(self.male1meas, CFG)
        except Exception as e:
            self.assertEqual(str(e),
                    "Number of CFG variables, 20, is incorrect.")

        # Invalid key.
        CFG = {'somersault': 0.0,
                'tilt': 0.0,
                'twist': np.pi/2,
                'PTsagittalFlexion': 0.0,
                'PTbending': 0.0,
                'TCspinalTorsion': 0.0,
                'TCsagittalSpinalFlexion': 0.0,
                'CA1extension': 0.0,
                'CA1adduction': 0.0,
                'CA1rotation': 0.0,
                'wrong': 0.01,
                'CB1extension': 0.0,
                'CB1abduction': np.pi/4,
                'CB1rotation': 0.0,
                'A1A2extension': 0.0,
                'PJ1extension': 0.0,
                'PJ1adduction': 0.0,
                'PK1extension': 0.0,
                'PK1abduction': 0.0,
                'J1J2flexion': 0.0,
                'K1K2flexion': 0.0,
                }
        self.assertRaises(Exception, hum.Human, self.male1meas, CFG)
        try:
            hum.Human(self.male1meas, CFG)
        except Exception as e:
            self.assertEqual(str(e),
                    "'wrong' is not a correct variable name.")

    # FIXME : This test is failing on Python 2.7 on Travis for an unknown
    # reason. As a work around, we skip the test when run on travis.
    @unittest.skipIf("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
                     "Skipping this test on Travis CI.")
    def test_validate_cfg(self):
        """Ensures that out-of-range values elicit a print, but no exception."""

        # Two values out of range.
        CFG = {'somersault': 0.0,
               'tilt': 0.0,
               'twist': 3.0 * np.pi,
               'PTsagittalFlexion': 0.0,
               'PTbending': 0.0,
               'TCspinalTorsion': 0.0,
               'TCsagittalSpinalFlexion': 0.0,
               'CA1extension': 0.0,
               'CA1adduction': 0.0,
               'CA1rotation': 0.0,
               'CB1extension': 0.0,
               'CB1abduction': np.pi / 4.0,
               'CB1rotation': 0.0,
               'A1A2extension': 0.0,
               'B1B2extension': 0.0,
               'PJ1extension': 0.0,
               'PJ1adduction': 0.0,
               'PK1extension': -10.0 * np.pi,
               'PK1abduction': 0.0,
               'J1J2flexion': 0.0,
               'K1K2flexion': 0.0,
               }
        desStr = ("Joint angle twist = 3.0 pi-rad is out of range. "
                "Must be between -1.0 and 1.0 pi-rad.\n"
                "Joint angle PK1extension = -10.0 pi-rad is out of range. "
                "Must be between -1.0 and 0.5 pi-rad.\n")

        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        hum.Human(self.male1meas, CFG)
        sys.stdout = old_stdout

        self.assertEquals(mystdout.getvalue(), desStr)

    def test_set_CFG(self):
        """Setting an individual CFG variable works. Also, error checks."""

        h = hum.Human(self.male1meas)
        h.set_CFG('K1K2flexion', np.pi * 0.2)

        testing.assert_almost_equal(h.CFG['K1K2flexion'], np.pi * 0.2)

        self.assertRaises(Exception, h.set_CFG, 'testing', 0.1)
        try:
            h.set_CFG('testing', 0.1)
        except Exception as e:
            self.assertEqual(str(e), "'testing' is not a valid name of a "
                    "configuration variable.")

        # Reset CFG.
        h = hum.Human(self.male1meas)
        # Inertia, etc, is affected correctly.
        h2 = hum.Human(self.male1meas)
        h2.set_CFG('somersault', np.pi / 2)
        # Relative inertia unchanged.
        testing.assert_allclose(h2.P.rel_inertia, h.P.rel_inertia)
        # Absolute inertia is changed.
        testing.assert_almost_equal(h2.inertia[0, 0], h.inertia[0, 0])
        testing.assert_almost_equal(h2.inertia[1, 1], h.inertia[2, 2])
        testing.assert_almost_equal(h2.inertia[2, 2], h.inertia[1, 1])

        h2.set_CFG('somersault', 0)
        h2.set_CFG('CA1extension', np.pi / 2)
        testing.assert_almost_equal(h2.A1.rel_inertia, h.A1.rel_inertia)
        testing.assert_almost_equal(h2.A2.rel_inertia, h.A2.rel_inertia)
        # y and z components are swapped.
        testing.assert_almost_equal(h2.A1.inertia[0, 0], h.A1.inertia[0, 0])
        testing.assert_almost_equal(h2.A1.inertia[1, 1], h.A1.inertia[2, 2])
        testing.assert_almost_equal(h2.A1.inertia[2, 2], h.A1.inertia[1, 1])
        testing.assert_almost_equal(h2.A2.inertia[0, 0], h.A2.inertia[0, 0])
        testing.assert_almost_equal(h2.A2.inertia[1, 1], h.A2.inertia[2, 2])
        testing.assert_almost_equal(h2.A2.inertia[2, 2], h.A2.inertia[1, 1])

        # Subtract out the arm, and see if the remaining inertia is good.
        # That is, the only part of the human's inertia tensor that changed
        # should be that which comes from the arm A.
        # Obtain inertia tensors about origin; the other option would be
        # h.center_of_mass, but this changes depending on A1A2extension.
        a1inertia_before = np.mat(inertia.parallel_axis( h.A1.inertia, h.A1.mass,
            h.A1.center_of_mass.T.tolist()[0]))
        a1inertia_after = np.mat(inertia.parallel_axis( h2.A1.inertia, h2.A1.mass,
            h2.A1.center_of_mass.T.tolist()[0]))
        a2inertia_before = np.mat(inertia.parallel_axis( h.A2.inertia, h.A2.mass,
            h.A2.center_of_mass.T.tolist()[0]))
        a2inertia_after = np.mat(inertia.parallel_axis( h2.A2.inertia, h2.A2.mass,
            h2.A2.center_of_mass.T.tolist()[0]))
        whole_inertia_before = np.mat(inertia.parallel_axis(h.inertia, h.mass,
            h.center_of_mass.T.tolist()[0]))
        whole_inertia_after = np.mat(inertia.parallel_axis(h2.inertia, h2.mass,
            h2.center_of_mass.T.tolist()[0]))
        testing.assert_allclose(
                whole_inertia_after - a1inertia_after - a2inertia_after,
                whole_inertia_before - a1inertia_before - a2inertia_before,
                atol=1e-15)

    def test_deprecated_CFG_names(self):
        """For v1.1.0, we changed a number of the configuration variable names.
        We still want to accept incorrect spellings for backwards
        compatibility. It should issue a deprecation warning
        # if you use the incorrect spelling but the code should run anyways.

        """
        deprecated_CFG_names = {'somersalt': 'somersault',
                'CA1elevation': 'CA1extension',
                'CB1elevation': 'CB1extension',
                'CA1abduction': 'CA1adduction',
                'A1A2flexion': 'A1A2extension',
                'B1B2flexion': 'B1B2extension',
                'TClateralSpinalFlexion': 'TCsagittalSpinalFlexion',
                'PJ1flexion': 'PJ1extension',
                'PK1flexion': 'PK1extension',
                'PJ1abduction': 'PJ1adduction',
                'PTfrontalFlexion': 'PTbending',
                }

        for old_name, new_name in deprecated_CFG_names.items():

            # set_CFG.
            h = hum.Human(self.male1meas)
            CFG_value = h.CFGbounds[h.CFGnames.index(new_name)][0]
            h.set_CFG(old_name, CFG_value)
            assert old_name not in h.CFG.keys()
            testing.assert_allclose(h.CFG[new_name], CFG_value)

            # set_CFG_dict.
            h = hum.Human(self.male1meas)
            CFG = copy.copy(h.CFG)
            CFG.pop(new_name)
            CFG[old_name] = CFG_value
            h.set_CFG_dict(CFG)
            assert old_name not in h.CFG.keys()
            assert new_name in h.CFG.keys()
            testing.assert_allclose(h.CFG[new_name], CFG_value)

        # Reading from file.
        cfg_path = os.path.join(os.path.split(__file__)[0],
                               'CFG_deprecated_names.txt')
        h = hum.Human(self.male1meas, cfg_path)
        for old_name, new_name in deprecated_CFG_names.items():
            assert old_name not in h.CFG.keys()
            assert new_name in h.CFG.keys()

    def test_set_CFG_dict(self):
        """Checks input errors and that the assignment occurs."""

        h = hum.Human(self.male1meas)

        h2 = hum.Human(self.male1meas)
        h2.set_CFG_dict(h.CFG)
        self.assertEqual(h2.CFG, h.CFG)

        CFG = copy.copy(h.CFG)
        CFG.pop('twist')
        self.assertRaises(Exception, h2.set_CFG_dict, CFG)
        try:
            h2.set_CFG_dict(CFG)
        except Exception as e:
            self.assertEqual(str(e),
                    "Number of CFG variables, 20, is incorrect.")

        CFG['testing'] = 0.5
        self.assertRaises(Exception, h2.set_CFG_dict, CFG)
        try:
            h2.set_CFG_dict(CFG)
        except Exception as e:
            self.assertEqual(str(e),
                    "'testing' is not a correct variable name.")

    def test_crazy_CFG_regression(self):
        """Puts the human in a crazy configuration (all angles are non-zero)
        and compares all 3 inertial properties.

        This is a regression test, not a physical verification.

        """
        CFG = {'somersault': 0.20 * np.pi,
               'tilt': 0.10 * np.pi,
               'twist': 0.30 * np.pi,
               'PTsagittalFlexion': 0.35 * np.pi,
               'PTbending': -0.18 * np.pi,
               'TCspinalTorsion': -0.08 * np.pi,
               'TCsagittalSpinalFlexion': 0.13 * np.pi,
               'CA1extension': -0.42 * np.pi,
               'CA1adduction': 0.19 * np.pi,
               'CA1rotation': 0.38 * np.pi,
               'CB1extension': -0.13 * np.pi,
               'CB1abduction': 0.32 * np.pi,
               'CB1rotation': 0.48 * np.pi,
               'A1A2extension': -0.45 * np.pi,
               'B1B2extension': -0.31 * np.pi,
               'PJ1extension': 0.26 * np.pi,
               'PJ1adduction': 0.37 * np.pi,
               'PK1extension': -0.21 * np.pi,
               'PK1abduction': 0.29 * np.pi,
               'J1J2flexion': 0.23 * np.pi,
               'K1K2flexion': 0.41 * np.pi,
               }
        h = hum.Human(self.male1meas, CFG)

        testing.assert_almost_equal(h.mass, 58.2004885884)

        expected_center_of_mass = np.matrix([[-0.04602766],
                                             [-0.17716871],
                                             [-0.05332974]])
        testing.assert_allclose(h.center_of_mass, expected_center_of_mass)

        expected_inertia = \
           np.matrix([[ 2.89276755,  0.43049026, -0.80508996],
                      [ 0.43049026,  4.37335248,  0.17229662],
                      [-0.80508996,  0.17229662,  4.13153827]])
        testing.assert_allclose(h.inertia, expected_inertia, atol=1e-6)

    def test_segment_pos(self):
        """Ensures that Segment.pos and Segment.end_pos return the correct
        quantities, for arms and legs. Regression test."""
        h = hum.Human(self.male1meas)
        testing.assert_almost_equal(h.A1.pos[2, 0], 0.472)
        testing.assert_almost_equal(h.A1.end_pos[2, 0], 0.2085)
        testing.assert_almost_equal(h.B1.pos[2, 0], 0.472)
        testing.assert_almost_equal(h.B1.end_pos[2, 0], 0.2085)

        testing.assert_almost_equal(h.J2.pos[2, 0], -0.424)
        testing.assert_almost_equal(h.J2.end_pos[2, 0], -1.02)

    def test_print_properties(self):
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        h = hum.Human(self.male1meas)
        h.print_properties()
        sys.stdout = old_stdout

        fname = os.path.join(os.path.split(__file__)[0],
                'human_print_des.txt')
        fid = open(fname, 'r')
        desStr = fid.read()
        fid.close()

        self.assertEquals(mystdout.getvalue(), desStr)

        # Use the __str__method.
        # It's just a fluke that we need to append an additional newline char.
        self.assertEquals(h.__str__() + '\n', desStr)

    def test_scale_human_by_mass(self):
        """User can scale human's mass, via meas input or API."""

        segment_names = ['head-neck', 'shoulders', 'thorax', 'abdomen-pelvis',
                'upper-arm', 'forearm', 'hand', 'thigh', 'lower-leg', 'foot']
        segmental_densities_des = {
            'Chandler': dict(zip(segment_names,
            [1056,  853,  853,  853, 1005, 1052, 1080, 1020, 1078, 1091])),
            'Dempster': dict(zip(segment_names,
            [1110, 1040,  920, 1010, 1070, 1130, 1160, 1050, 1090, 1100])),
            'Clauser': dict(zip(segment_names,
            [1070, 1019, 1019, 1019, 1056, 1089, 1109, 1044, 1085, 1084])),
            }

        # For comparison, unscaled densities.
        h = hum.Human(self.male1meas)

        # From measurement file, scaling mass though.
        measPath = os.path.join(os.path.split(__file__)[0],
                'male1_scale.txt')
        h2 = hum.Human(measPath)

        # Make sure the mass is 100.
        self.assertEqual(h2.mass, 100)

        # Make sure all density values are scaled appropriately.
        factor = h2.mass / h.mass

        # Make sure densities are scaled correctly.
        for key, val in h.segmental_densities.items():
            for seg, dens in val.items():
                self.assertEquals(dens,
                        segmental_densities_des[key][seg] * factor)

        # Check a few individual segments and solids.
        testing.assert_almost_equal(h2.K1.mass, h.K1.mass * factor)
        testing.assert_almost_equal(h2.A2.solids[0].mass, h.A2.solids[0].mass * factor)

        # Check center of mass is not modified.
        testing.assert_almost_equal(h2.center_of_mass[2], h.center_of_mass[2])
        # Check that inertia is modified.
        testing.assert_allclose(h2.inertia, h.inertia * factor, atol=1e-15)
        testing.assert_allclose(h2.B1.solids[0].inertia, h.B1.solids[0].inertia
                * factor)

        # Scale again.
        h2.scale_human_by_mass(30)
        testing.assert_almost_equal(h2.mass, 30)

        factor2 = h2.mass / 100
        # Make sure densities are scaled correctly.
        for key, val in h2.segmental_densities.items():
            for seg, dens in val.items():
                self.assertEquals(dens,
                        segmental_densities_des[key][seg] * factor * factor2)

    def test_read_measurements(self):
        # -- Measurement input file errors.
        measPath = os.path.join(os.path.split(__file__)[0],
                'male1_badkey.txt')
        self.assertRaises(ValueError, hum.Human, measPath)
        try:
            h = hum.Human(measPath)
        except ValueError as e:
            self.assertEqual(str(e), "Variable LsLL is not "
                    "valid name for a measurement.")

        measPath = os.path.join(os.path.split(__file__)[0],
                'male1_badval.txt')
        self.assertRaises(ValueError, hum.Human, measPath)
        try:
            h = hum.Human(measPath)
        except ValueError as e:
            self.assertEqual(str(e),
                    "Variable Ls1L has inappropriate value.")

        measPath = os.path.join(os.path.split(__file__)[0],
                'male1_mcf.txt')
        self.assertRaises(Exception, hum.Human, measPath)
        try:
            h = hum.Human(measPath)
        except Exception as e:
            self.assertEqual(str(e),
                    "Variable measurementconversionfactor not provided "
                    "or is 0. Set as 1 if measurements are given in meters.")

        measPath = os.path.join(os.path.split(__file__)[0],
                'male1_missingkey.txt')
        self.assertRaises(Exception, hum.Human, measPath)
        try:
            h = hum.Human(measPath)
        except Exception as e:
            self.assertEqual(str(e), "There should be 95 "
                    "measurements, but 94 were found.")

    def test_write_measurements(self):
        """Writes a valid YAML file that can be read back in."""

        path = os.path.join(os.path.split(__file__)[0],
                'meas_output.txt')
        pathDes = os.path.join(os.path.split(__file__)[0],
                'meas_output_des.txt')

        h = hum.Human(self.male1meas)
        h.write_measurements(path)

        # Output is correct.
        self.assertEqual(open(path, 'r').read(), open(pathDes, 'r').read())

        # Works as input.
        h2 = hum.Human(path)
        self.assertEqual(h2.meas, h.meas)

        os.remove(path)

    def test_write_meas_for_ISEG(self):
        """Ensures ISEG input is written correctly."""

        path = os.path.join(os.path.split(__file__)[0],
                'meas_iseg.txt')
        pathDes = os.path.join(os.path.split(__file__)[0],
                'meas_iseg_des.txt')

        h = hum.Human(self.male1meas)
        h.write_meas_for_ISEG(path)

        self.assertEqual(open(path, 'r').read(), open(pathDes, 'r').read())

        os.remove(path)

    def test_read_CFG(self):
        """Particularly checks input errors."""

        # Unrecognized variable.
        cfgPath = os.path.join(os.path.split(__file__)[0],
                'CFG_badkey.txt')
        self.assertRaises(ValueError, hum.Human, self.male1meas, cfgPath)
        try:
            hum.Human(self.male1meas, cfgPath)
        except ValueError as e:
            self.assertEqual(str(e),
                    "'invalid' is not a correct variable name.")

        # Too few inputs.
        cfgPath = os.path.join(os.path.split(__file__)[0],
                'CFG_missingkey.txt')
        self.assertRaises(ValueError, hum.Human, self.male1meas, cfgPath)
        try:
            hum.Human(self.male1meas, cfgPath)
        except ValueError as e:
            self.assertEqual(str(e), "Number of CFG variables, 20, is "
                    "incorrect.")

        # No value for a key.
        cfgPath = os.path.join(os.path.split(__file__)[0],
                'CFG_badval.txt')
        self.assertRaises(ValueError, hum.Human, self.male1meas, cfgPath)
        try:
            hum.Human(self.male1meas, cfgPath)
        except ValueError as e:
            self.assertEqual(str(e),
                    "Variable PTsagittalFlexion has no value.")

    def test_write_CFG(self):
        """Writes a valid YAML file that can be read back in."""

        CFG = {'somersault': 0.0,
                'tilt': 0.0,
                'twist': np.pi/2,
                'PTsagittalFlexion': 0.0,
                'PTbending': 0.0,
                'TCspinalTorsion': 0.0,
                'TCsagittalSpinalFlexion': 0.0,
                'CA1extension': 0.0,
                'CA1adduction': 0.0,
                'CA1rotation': 0.0,
                'CB1extension': 0.0,
                'CB1abduction': np.pi/4,
                'CB1rotation': 0.0,
                'A1A2extension': 0.0,
                'B1B2extension': 0.0,
                'PJ1extension': 0.0,
                'PJ1adduction': 0.0,
                'PK1extension': 0.0,
                'PK1abduction': 0.0,
                'J1J2flexion': 0.0,
                'K1K2flexion': 0.0,
                }
        path = os.path.join(os.path.split(__file__)[0],
                'CFG_output.txt')
        h = hum.Human(self.male1meas, CFG)
        h.write_CFG(path)

        pathDes = os.path.join(os.path.split(__file__)[0],
                'CFG_output_des.txt')
        self.assertEqual(open(path, 'r').read(), open(pathDes, 'r').read())

        # Can use the cfg output as input.
        h = hum.Human(self.male1meas, path)
        self.assertEqual(h.CFG, CFG)

        os.remove(path)

    def test_translate_coord_sys(self):
        """Just translates once and makes sure only COM changes."""
        # TODO this method has been hidden for v1.0 release.
        h = hum.Human(self.male1meas)
        h2 = hum.Human(self.male1meas)
        h2._translate_coord_sys([1, 2, 3])

        testing.assert_almost_equal(h2.mass, h.mass)
        testing.assert_allclose(h2.center_of_mass,
                h.center_of_mass + np.array([[1], [2], [3]]))
        testing.assert_allclose(h2.inertia, h.inertia, atol=1e-15)

    def test_rotate_coord_sys(self):
        # TODO Check out properties again.
        h = hum.Human(self.male1meas)
        # TODO this method has been hidden for v1.0 release.
        pass

    def test_transform_coord_sys(self):
        # TODO this can be simple.
        # TODO this method has been hidden for v1.0 release.
        pass

    def test_combine_inertia(self):
        """Tries input errors and checks output against some hand
        calculations."""
        # Input errors.
        h = hum.Human(self.male1meas, symmetric=True)

        # Nonsensical input.
        with self.assertRaises(Exception) as e:
            h.combine_inertia([''])
        self.assertEquals(str(e.exception), "The string '' does not "
                "identify a segment or solid of the human.")

        with self.assertRaises(Exception) as e:
            h.combine_inertia()

        with self.assertRaises(Exception) as e:
            h.combine_inertia([])
        self.assertEquals(str(e.exception), "Empty input.")

        # List a string that doesn't identify a segment or solid.
        with self.assertRaises(Exception) as e:
            h.combine_inertia(['abracadabra'])
        self.assertEquals(str(e.exception), "The string 'abracadabra' does "
                "not identify a segment or solid of the human.")

        # List a solid more than once.
        with self.assertRaises(Exception) as e:
            h.combine_inertia(['j4', 'a0', 's0', 'a0', 'A2'])
        self.assertEquals(str(e.exception), "An object is listed more than "
                "once. A solid/segment can only be listed once.")

        # List solid and its parent segment.
        with self.assertRaises(Exception) as e:
            h.combine_inertia(['k2', 'K1'])
        self.assertEquals(str(e.exception), "A solid k2 and its parent "
                "segment K1 have both been given as inputs. This duplicates "
                "that solid's contribution.")

        with self.assertRaises(Exception) as e:
            h.combine_inertia(['B2', 'b6'])

        # The following depends on the fact that we've imposed symmetry.
        # 'c' for combined.
        c_mass, c_com, c_inertia = h.combine_inertia(['a0', 'b0'])
        # Using _a_solids instead of A1.solids because we've reversed the order
        # of the solids in the segments.
        c_mass_des = h._a_solids[0].mass + h._b_solids[0].mass
        self.assertEquals(c_mass, c_mass_des)
        testing.assert_allclose(c_com,
                (h._a_solids[0].mass * h._a_solids[0].center_of_mass +
                    h._b_solids[0].mass * h._b_solids[0].center_of_mass) /
                c_mass_des)
        # Given symmetry, com's x-component should be zero.
        self.assertEquals(c_com[0, 0], 0.0)
        # ... z-component is the same as for individual solids.
        self.assertEquals(c_com[2, 0], h._a_solids[0].center_of_mass[2, 0])
        # The 2 comes from symmetry; then we have parallel axis theorem.
        self.assertEquals(c_inertia[0, 0], 2 * h._a_solids[0].inertia[0, 0])
        parallel_term = \
                h._a_solids[0].mass * h._a_solids[0].center_of_mass[0, 0]**2
        Iyy_des = 2.0 * (h._a_solids[0].inertia[1, 1] + parallel_term)
        self.assertEquals(c_inertia[1, 1], Iyy_des)
        Izz_des = 2.0 * (h._a_solids[0].inertia[2, 2] + parallel_term)
        self.assertEquals(c_inertia[2, 2], Izz_des)
        self.assertEquals(c_inertia[0, 1], 0.0)
        self.assertEquals(c_inertia[0, 2], 0.0)
        self.assertEquals(c_inertia[1, 2], 0.0)

        c_mass = None
        c_com = None
        c_inertia = None

        # Try segments.
        c_mass, c_com, c_inertia = h.combine_inertia(['A2', 'B2'])
        c_mass_des = h.A2.mass + h.B2.mass
        self.assertEquals(c_mass, c_mass_des)
        self.assertEquals(c_com[0, 0], 0.0)
        testing.assert_almost_equal(c_com[2, 0], h.B2.center_of_mass[2, 0])
        self.assertEquals(c_inertia[0, 0], 2 * h.A2.inertia[0, 0])
        parallel_term = h.A2.mass * h.A2.center_of_mass[0, 0]**2
        Iyy_des = 2.0 * (h.A2.inertia[1, 1] + parallel_term)
        self.assertEquals(c_inertia[1, 1], Iyy_des)
        Izz_des = 2.0 * (h.A2.inertia[2, 2] + parallel_term)
        self.assertEquals(c_inertia[2, 2], Izz_des)

        self.assertEquals(c_inertia[0, 1], 0.0)
        testing.assert_almost_equal(c_inertia[0, 2], 0.0)
        self.assertEquals(c_inertia[1, 2], 0.0)

    def test_inertia_transformed(self):
        """Tests the functionality of getting an inertia tensor about a
        different point and in a different frame.

        """
        h = hum.Human(self.male1meas)
        h.set_CFG('somersault', np.pi * 0.25)
        inertia_pre = h.inertia
        m = h.mass

        # A simple change in position.
        d = 15  # meters
        inertia_post = h.inertia_transformed(
                pos=[d + h.center_of_mass[0, 0],
                    h.center_of_mass[1, 0],
                    h.center_of_mass[2, 0]])
        offset = m * d**2
        # Moments of inertia.
        testing.assert_almost_equal(inertia_post[0, 0], inertia_pre[0, 0])
        testing.assert_almost_equal(
                inertia_post[1, 1], inertia_pre[1, 1] + offset)
        testing.assert_almost_equal(
                inertia_post[2, 2], inertia_pre[2, 2] + offset)
        # Products of inertia.
        testing.assert_almost_equal(inertia_post[0, 1], inertia_pre[0, 1])
        testing.assert_almost_equal(inertia_post[0, 2], inertia_pre[0, 2])
        testing.assert_almost_equal(inertia_post[1, 2], inertia_pre[1, 2])
        # Symmetry is preserved.
        testing.assert_almost_equal(inertia_post[1, 0], inertia_post[0, 1])
        testing.assert_almost_equal(inertia_post[2, 0], inertia_post[0, 2])
        testing.assert_almost_equal(inertia_post[2, 1], inertia_post[1, 2])
        inertia_post = None

        # A more complicated change in position.
        inertia_post = h.inertia_transformed(
                pos=[d + h.center_of_mass[0, 0],
                     h.center_of_mass[1, 0],
                     2 * d + h.center_of_mass[2, 0]])
        offset2 = m * 4 * d**2
        # Moments of inertia.
        testing.assert_almost_equal(
                inertia_post[0, 0], inertia_pre[0, 0] + offset2)
        testing.assert_almost_equal(
                inertia_post[1, 1], inertia_pre[1, 1] + offset + offset2)
        testing.assert_almost_equal(
                inertia_post[2, 2], inertia_pre[2, 2] + offset)
        # Products of inertia.
        testing.assert_almost_equal(inertia_post[0, 1], inertia_pre[0, 1])
        testing.assert_almost_equal(
                inertia_post[0, 2], inertia_pre[0, 2] - m * 2 * d * d)
        testing.assert_almost_equal(inertia_post[1, 2], inertia_pre[1, 2])
        # Symmetry is preserved.
        testing.assert_almost_equal(inertia_post[1, 0], inertia_post[0, 1])
        testing.assert_almost_equal(inertia_post[2, 0], inertia_post[0, 2])
        testing.assert_almost_equal(inertia_post[2, 1], inertia_post[1, 2])
        inertia_post = None

        # A combined change in position and basis. Shift in x and rotation
        # about X.
        inertia_post = h.inertia_transformed(
                pos=[d + h.center_of_mass[0, 0],
                         h.center_of_mass[1, 0],
                         h.center_of_mass[2, 0]],
                rotmat=inertia.rotate_space_123((0.5 * np.pi, 0.0, 0.0)))
        # Moments of inertia.
        testing.assert_almost_equal(inertia_post[0, 0], inertia_pre[0, 0])
        testing.assert_almost_equal(
                inertia_post[1, 1], inertia_pre[2, 2] + offset)
        testing.assert_almost_equal(
                inertia_post[2, 2], inertia_pre[1, 1] + offset)
        # Products of inertia.
        testing.assert_almost_equal(inertia_post[0, 1], inertia_pre[0, 2])
        testing.assert_almost_equal(inertia_post[0, 2], -inertia_pre[0, 1])
        testing.assert_almost_equal(inertia_post[1, 2], -inertia_pre[1, 2])
        # Symmetry is preserved.
        testing.assert_almost_equal(inertia_post[1, 0], inertia_post[0, 1])
        testing.assert_almost_equal(inertia_post[2, 0], inertia_post[0, 2])
        testing.assert_almost_equal(inertia_post[2, 1], inertia_post[1, 2])
        inertia_post = None

        # Make sure the direction of the rotation matrix is correct.
        # Inertia tensor of 2 point mass in the y-z plane; at (2, 1) and (-2,
        # -1), both with mass 1.
        # Rotating a positive atan(1/2) should give a zero xy product of
        # inertia if the rotation matrix is what we think it is.
        inertia_ptmass = np.mat(np.zeros((3, 3)))
        inertia_ptmass[0, 0] = 2 * 1 * (2**2 + 1**1)
        inertia_ptmass[1, 1] = 2 * 1 * (1**1)
        inertia_ptmass[2, 2] = 2 * 1 * (2**2)
        Iyz = 2 * 1 * (2 * 1)
        inertia_ptmass[1, 2] = -Iyz
        inertia_ptmass[2, 1] = -Iyz # symmetry
        angle = np.arctan(1.0 / 2.0)
        # This returns R from va = R * vb:
        rotmat = inertia.rotate_space_123((angle, 0.0, 0.0))
        # Here is a little cheating to test out the method itself with our own
        # center of mass and inertia:
        h._mass = 1
        h._center_of_mass = np.array([[0], [0], [0]])
        h._inertia = inertia_ptmass
        # This expects R from va = R * vb:
        inertia_post = h.inertia_transformed(rotmat=rotmat)
        # Moments of inertia.
        testing.assert_almost_equal(inertia_post[0, 0], inertia_ptmass[0, 0])
        testing.assert_almost_equal(inertia_post[1, 1], 0.0)
        testing.assert_almost_equal(inertia_post[2, 2], inertia_ptmass[0, 0])
        # Products of inertia.
        testing.assert_almost_equal(inertia_post[0, 1], 0.0)
        testing.assert_almost_equal(inertia_post[0, 2], 0.0)
        testing.assert_almost_equal(inertia_post[1, 2], 0.0)
        # Symmetry is preserved.
        testing.assert_almost_equal(inertia_post[1, 0], inertia_post[0, 1])
        testing.assert_almost_equal(inertia_post[2, 0], inertia_post[0, 2])
        testing.assert_almost_equal(inertia_post[2, 1], inertia_post[1, 2])
        inertia_post = None

    def test_lower_torso_rotations(self):
        """Yeadon specifies Euler 1-2-3 rotations (body fixed 1-2-3). For the
        lower torso, this is somersault-tilt-twist relative to the inertial
        reference frame."""

        h = hum.Human(self.male1meas)

        # specify some arbitrary angles
        somersault = pi / 5.0
        tilt = pi / 10.0
        twist = pi / 14.0

        h.set_CFG('somersault', somersault)
        h.set_CFG('tilt', tilt)
        h.set_CFG('twist', twist)

        # Now manually calculate the rotation matrix for Euler 1-2-3 using the
        # inerita.euler_rotation function. Yeadon starts with the I frame
        # (inertial) and rotates the F frame (attached to P) through the body
        # fixed 123 angles.
        # NOTE inertia.euler_rotation has been removed.

        R = inertia.euler_123((somersault, tilt, twist))

        testing.assert_almost_equal(h.P.rot_mat, R)

        # Yeadon presents this rotation matrix on page 3 of his first paper,
        # but it has two errors. The entry Sfi[1, 0] should be `+ sin(phi)`
        # instead of `+ sin(theta)` and the entry Sfi[1, 1] should be
        # `cos(phi)cos(psi) -` instead of `cos(theta)cos(psi) -`. The following
        # is the matrix with the error corrected. His definition of the
        # rotation matrix is:

        # v_i = Sfi * v_f

        # which is the same as the definition in euler_123

        phi = somersault
        theta = tilt
        psi = twist

        cph = np.cos(phi)
        sph = np.sin(phi)
        cth = np.cos(theta)
        sth = np.sin(theta)
        cps = np.cos(psi)
        sps = np.sin(psi)

        Sfi = np.mat(
            [[cth * cps, -cth * sps, sth],
             [cph * sps + sph * sth * cps,
              cph * cps - sph * sth * sps,
              -sph * cth],
             [sph * sps - cph * sth * cps,
              sph * cps + cph * sth * sps,
              cph * cth]])

        testing.assert_almost_equal(h.P.rot_mat, Sfi)

    def test_leg_rotations(self):

        h = hum.Human(self.male1meas)

        elevation = pi / 10.05
        abduction = pi / 68.4

        h.set_CFG('PJ1extension', elevation)
        h.set_CFG('PJ1adduction', -abduction) # should be neg

        RJ = inertia.euler_123((elevation, -abduction, 0.0))

        testing.assert_allclose(h.J1.rot_mat, RJ)

        h.set_CFG('PK1extension', elevation)
        h.set_CFG('PK1abduction', abduction)

        RK = inertia.euler_123((elevation, abduction, 0.0))

        testing.assert_allclose(h.K1.rot_mat, RK)

        flexion = pi / 9.6

        R2 = inertia.euler_123((flexion, 0.0, 0.0))

        h.set_CFG('J1J2flexion', flexion)
        h.set_CFG('K1K2flexion', flexion)

        testing.assert_allclose(h.J2.rot_mat, RJ * R2)
        testing.assert_allclose(h.K2.rot_mat, RK * R2)

        somersault = pi / 5.0
        tilt = pi / 10.0
        twist = pi / 14.0

        h.set_CFG('somersault', somersault)
        h.set_CFG('tilt', tilt)
        h.set_CFG('twist', twist)

        R3 = inertia.euler_123((somersault, tilt, twist))

        # v_p = R3 * v_i
        # v_j1 = R * v_p
        # v_j2 = R2 * v_j1
        # v_j2 = R2 * R * R3 * v_i
        # v_i = R3^T * R^T * R2^T * v_j2

        testing.assert_allclose(h.J2.rot_mat, R3 * RJ * R2)
        testing.assert_allclose(h.K2.rot_mat, R3 * RK * R2)

    def test_arm_rotations(self):
        """Yeadon specifies Euler 1-2-3 rotations (body fixed 1-2-3). For the
        left arm, A, this is elevation-abduction-rotation relative to the C
        body (shoulders/head)."""

        h = hum.Human(self.male1meas)

        elevation = pi / 5.0
        abduction = pi - pi / 10.0
        rotation = pi / 14.0

        h.set_CFG('CA1extension', elevation)
        h.set_CFG('CA1adduction', abduction)
        h.set_CFG('CA1rotation', rotation)

        # now we should manually calculate the rotation matrix for Euler 1-2-3
        # Yeadon starts with the C frame and rotates the F frame
        # (attached to P) through the body fixed 123 angles.

        R = inertia.euler_123((elevation, abduction, rotation))

        testing.assert_allclose(h.A1.rot_mat, R)

        # now rotate the elbow

        flexion = - pi / 4.3

        R2 = inertia.euler_123((flexion, 0.0, 0.0))

        h.set_CFG('A1A2extension', flexion)

        testing.assert_allclose(h.A2.rot_mat, R * R2)

        # right arm

        elevation = pi / 5.0
        abduction = pi - pi / 10.0
        rotation = pi / 14.0

        h.set_CFG('CB1extension', elevation)
        h.set_CFG('CB1abduction', abduction)
        h.set_CFG('CB1rotation', rotation)

        # now we should manually calculate the rotation matrix for Euler 1-2-3
        # Yeadon starts with the C frame and rotates the F frame
        # (attached to P) through the body fixed 123 angles.

        R = inertia.euler_123((elevation, abduction, rotation))

        testing.assert_allclose(h.B1.rot_mat, R)

        # now rotate the elbow
        # v_a1 = R * v_c and v_a2 = R2 * v_a1
        # v_a2 = R2 * R * v_c
        # v_c =  R^T * R2^T * v_a2

        flexion = -pi / 6.98

        R2 = inertia.euler_123((flexion, 0.0, 0.0))

        h.set_CFG('B1B2extension', flexion)

        testing.assert_allclose(h.B2.rot_mat, R * R2)

    def test_rotation_chain(self):
        """This tests all of the rotations in the whole body."""

        h = hum.Human(self.male1meas)

        def angle():
            """Returns a random angle between -2*pi and 2*pi."""
            return 2 * pi * np.random.uniform(-1, 1)

        # rotate the pelvis (P) relative to the inertial frame (I)

        somersault = angle()
        tilt = angle()
        twist = angle()

        h.set_CFG('somersault', somersault)
        h.set_CFG('tilt', tilt)
        h.set_CFG('twist', twist)

        P_R_I = inertia.euler_123((somersault, tilt, twist))

        testing.assert_allclose(h.P.rot_mat, P_R_I)

        # rotate the thorax (T) relative to the pelvis (P)

        sagflexion = angle()
        bending = angle()

        h.set_CFG('PTsagittalFlexion', sagflexion)
        h.set_CFG('PTbending', bending)

        T_R_P = inertia.euler_123((sagflexion, bending, 0.0))

        T_R_I = P_R_I * T_R_P

        testing.assert_allclose(h.T.rot_mat, T_R_I)

        # rotate the chest relative to the thorax

        spinalflexion = angle()
        torsion = angle()

        h.set_CFG('TCsagittalSpinalFlexion', spinalflexion)
        h.set_CFG('TCspinalTorsion', torsion)

        C_R_T = inertia.euler_123((spinalflexion, 0.0, torsion))

        C_R_I = T_R_I * C_R_T

        testing.assert_allclose(h.C.rot_mat, C_R_I)

        # rotate the left upper arm relative to the chest

        extension = angle()
        abduction = angle() # neg
        rotation = angle()

        h.set_CFG('CA1extension', extension)
        h.set_CFG('CA1adduction', abduction)
        h.set_CFG('CA1rotation', rotation)

        A1_R_C = inertia.euler_123((extension, abduction, rotation))

        A1_R_I = C_R_I * A1_R_C

        testing.assert_allclose(h.A1.rot_mat, A1_R_I)

        # rotate the left lower arm relative to the left upper arm

        extension = angle() #neg

        h.set_CFG('A1A2extension', extension)

        A2_R_A1 = inertia.euler_123((extension, 0.0, 0.0))

        A2_R_I = A1_R_I * A2_R_A1

        testing.assert_allclose(h.A2.rot_mat, A2_R_I)

        # rotate the right upper arm relative to the chest

        extension = angle()
        abduction = angle()
        rotation = angle()

        h.set_CFG('CB1extension', extension)
        h.set_CFG('CB1abduction', abduction)
        h.set_CFG('CB1rotation', rotation)

        B1_R_C = inertia.euler_123((extension, abduction, rotation))

        B1_R_I = C_R_I * B1_R_C

        testing.assert_allclose(h.B1.rot_mat, B1_R_I)

        # rotate the left lower arm relative to the left upper arm

        extension = angle() # neg

        h.set_CFG('B1B2extension', extension)

        B2_R_B1 = inertia.euler_123((extension, 0.0, 0.0))

        B2_R_I = B1_R_I * B2_R_B1

        testing.assert_allclose(h.B2.rot_mat, B2_R_I)

        # legs

        # left leg

        elevation = angle()
        abduction = angle() # neg

        h.set_CFG('PJ1extension', elevation)
        h.set_CFG('PJ1adduction', abduction) # should be neg

        J1_R_P = inertia.euler_123((elevation, abduction, 0.0))

        J1_R_I = P_R_I * J1_R_P

        testing.assert_allclose(h.J1.rot_mat, J1_R_I)

        flexion = angle()

        h.set_CFG('J1J2flexion', flexion)

        J2_R_J1 = inertia.euler_123((flexion, 0.0, 0.0))

        J2_R_I = J1_R_I * J2_R_J1

        testing.assert_allclose(h.J2.rot_mat, J2_R_I)

        # right leg

        elevation = angle()
        abduction = angle()

        h.set_CFG('PK1extension', elevation)
        h.set_CFG('PK1abduction', abduction) # should be neg

        K1_R_P = inertia.euler_123((elevation, abduction, 0.0))

        K1_R_I = P_R_I * K1_R_P

        testing.assert_allclose(h.K1.rot_mat, K1_R_I)

        flexion = angle()

        h.set_CFG('K1K2flexion', flexion)

        K2_R_K1 = inertia.euler_123((flexion, 0.0, 0.0))

        K2_R_I = K1_R_I * K2_R_K1

        testing.assert_allclose(h.K2.rot_mat, K2_R_I)

# TODO compare ISEG output to our output.

# TODO try out a program flow: make sure we do all necessary updates after
# construction, say when we change a joint angle, etc. CANNOT just change
# measurements on the fly.
# TODO it's possible that by averaging the measurements, we're coming up
# with more false stadia than we would otherwise, but if we don't average
# the measurements, how do we draw?



#!/usr/bin/env python

# Use Python3 integer division rules.
from __future__ import division

import sys
import os
import unittest
import warnings

from numpy import testing, pi, array, matrix, sin, cos, zeros, array, mat, \
        arctan

from yeadon.solid import Stadium, Solid, StadiumSolid
from yeadon import inertia

warnings.filterwarnings('ignore', category=DeprecationWarning)


class StadiumSolidCheck(unittest.TestCase):
    """To check the formulae in yeadon.solid against those in the Yeadon1990-ii
    paper.

    """
    def __init__(self, density, thick0, rad0, thick1, rad1, height):
        self.D = density
        self.t0 = thick0
        self.r0 = rad0
        self.t1 = thick1
        self.r1 = rad1
        self.h = height
        self.a = (self.r1 - self.r0) / self.r0
        self.b = (self.t1 - self.t0) / self.t0

    def mass(s):
        return s.D * s.h * s.r0 * (
                4.0 * s.t0 * s._F1(s.a, s.b) +
                pi * s.r0 * s._F1(s.a, s.a)
                )

    def mass_center(s):
        return s.D * s.h**2 * (
                4.0 * s.r0 * s.t0 * s._F2(s.a, s.b) +
                pi * s.r0**2 * s._F2(s.a, s.a)
                ) / s.mass()

    def inertia_zz(s):
        """About center of mass."""
        return s.D * s.h * (
                4.0 * s.r0 * s.t0**3 * s._F4(s.a, s.b) / 3.0 +
                pi * s.r0**2 * s.t0**2 * s._F5(s.a, s.b) +
                4.0 * s.r0**3 * s.t0 * s._F4(s.b, s.a) +
                pi * s.r0**4 * s._F4(s.a, s.a) / 2.0
                )

    def inertia_yy(s):
        """About center of mass."""
        Jy_integral = (4.0 * s.r0 * s.t0**3 * s._F4(s.a, s.b) / 3.0 +
                pi * s.r0**2 * s.t0**2 * s._F5(s.a, s.b) +
                8.0 * s.r0**3 * s.t0 * s._F4(s.b, s.a) / 3.0 +
                pi * s.r0**4 * s._F4(s.a, s.a) / 4.0
                )
        z2A_integral = (4.0 * s.r0 * s.t0 * s._F3(s.a, s.b) +
                pi * s.r0**2 * s._F3(s.a, s.a))
        about_origin = s.D * s.h * Jy_integral + s.D * s.h**3 * z2A_integral
        return about_origin - s.mass() * s.mass_center()**2

    def inertia_xx(s):
        """About center of mass."""
        Jz_integral = (4.0 * s.r0 * s.t0**3 * s._F4(s.a, s.b) / 3.0 +
                pi * s.r0**4 * s._F4(s.a, s.a) / 4.0)
        z2A_integral = (4.0 * s.r0 * s.t0 * s._F3(s.a, s.b) +
                pi * s.r0**2 * s._F3(s.a, s.a))
        about_origin = s.D * s.h * Jz_integral + s.D * s.h**3 * z2A_integral
        return about_origin - s.mass() * s.mass_center()**2

    @staticmethod
    def _F1(a, b):
        return 1.0 + (a + b)/2.0 + a*b/3.0

    @staticmethod
    def _F2(a, b):
        return 0.5 + (a + b)/3.0 + a*b/4.0

    @staticmethod
    def _F3(a, b):
        return 1/3.0 + (a + b)/4.0 + a*b/5.0

    @staticmethod
    def _F4(a, b):
        return (1.0 + (a + 3.0*b)/2.0 + (3.0*a*b + 3.0*b**2)/3.0 +
                (3.0*a*b**2 + b**3)/4.0 + a*b**3/5.0)

    @staticmethod
    def _F5(a, b):
        return (1.0 + (2.0*a + 2.0*b)/2.0 + (a**2 + 4.0*a*b + b**2)/3.0 +
                2.0*a*b*(a + b)/4.0 + a**2 * b**2 / 5.0)


# define some useful functions for 2D stadia
def radius_from_perimeter_width(perimeter, width):
    """Returns the radius of the stadium given the perimeter and width."""
    return (perimeter - 2.0 * width) / (2 * pi - 4)

def thickness_from_perimeter_width(perimeter, width):
    """Returns the thickness of the stadium given the perimeter and
    width."""
    return 0.5 * width - radius_from_perimeter_width(perimeter, width)

def radius_from_depth(depth):
    return depth / 2.0

def thickness_from_depth_width(depth, width):
    return (width - depth) / 2.0

def perimeter_from_depth_width(depth, width):
    return 2 * (width - depth) + pi * depth

class TestStadium(unittest.TestCase):

    def test_init(self):
        # perimeter and width
        perimeter = 2.5
        width = 1.0
        pw = Stadium('La6: knuckles', 'perimwidth', perimeter, width)
        assert pw.label == 'La6: knuckles'
        assert pw.alignment == 'ML'
        testing.assert_almost_equal(pw.perimeter, perimeter)
        testing.assert_almost_equal(pw.width, width)
        testing.assert_almost_equal(pw.radius,
                radius_from_perimeter_width(perimeter, width))
        testing.assert_almost_equal(pw.thickness,
                thickness_from_perimeter_width(perimeter, width))

        # depth and width
        depth = 1.0
        width = 5.0
        dw = Stadium('La6: knuckles', 'depthwidth', depth, width, 'AP')
        assert dw.label == 'La6: knuckles'
        assert dw.alignment == 'AP'
        testing.assert_almost_equal(dw.width, width)
        testing.assert_almost_equal(dw.radius, radius_from_depth(depth))
        testing.assert_almost_equal(dw.thickness,
                thickness_from_depth_width(depth, width))
        testing.assert_almost_equal(dw.perimeter,
                perimeter_from_depth_width(depth, width))

        # perim
        perimeter = 5.0
        p = Stadium('Lk2: mid-thigh', 'perimeter', perimeter)
        assert p.label == 'Lk2: mid-thigh'
        assert p.alignment == 'ML'
        testing.assert_almost_equal(p.perimeter, perimeter)
        testing.assert_almost_equal(p.width, perimeter / pi)
        testing.assert_almost_equal(p.thickness, 0.0)
        testing.assert_almost_equal(p.radius, perimeter / 2.0 / pi)

        # radius
        radius = 1.0
        r = Stadium('Lk2: mid-thigh', 'radius', radius)
        assert r.label == 'Lk2: mid-thigh'
        assert r.alignment == 'ML'
        testing.assert_almost_equal(r.radius, radius)
        testing.assert_almost_equal(r.perimeter, 2.0 * pi * radius)
        testing.assert_almost_equal(r.thickness, 0.0)
        testing.assert_almost_equal(r.width, 2 * radius)

        # thickness and radius
        thickness = 10.0
        radius = 1.0
        tr = Stadium('Lk2: mid-thigh', 'thicknessradius', thickness, radius)
        assert tr.label == 'Lk2: mid-thigh'
        assert tr.alignment == 'ML'
        testing.assert_almost_equal(tr.radius, radius)
        testing.assert_almost_equal(tr.thickness, thickness)
        testing.assert_almost_equal(tr.perimeter, 2.0 * pi * radius + 4.0 *
                thickness)
        testing.assert_almost_equal(tr.width, 2 * radius + 2.0 * thickness)

    def test_invalid_stadium(self):
        """Tests that if a stadium is defined in such a way that it is invalid
        (negative radius or negative thickness), the correct action is taken.

        """
        # TODO Redirecting stdout is not working.
        actual_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            stad = Stadium('Lb1: mid-arm', 'perimwidth', 1.9, 1.0)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            assert "incorrectly" in str(w[-1])
            testing.assert_almost_equal(stad.perimeter, 1.9)
            testing.assert_almost_equal(stad.radius, 1.9 / (2.0 * pi))
            testing.assert_almost_equal(stad.thickness, 0.0)
            testing.assert_almost_equal(stad.width, 1.9 / pi)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            stad = Stadium('Lb1: mid-arm', 'perimwidth', 3.15, 1.0)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            testing.assert_almost_equal(stad.perimeter, 3.15)
            testing.assert_almost_equal(stad.radius, 3.15 / (2.0 * pi))
            testing.assert_almost_equal(stad.thickness, 0.0)
            testing.assert_almost_equal(stad.width, 3.15 / pi)

        width = 1.0
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            depth = (1.9 - 2.0 * width) / (pi - 2.0)
            stad = Stadium('Lb1: mid-arm', 'depthwidth', depth, width)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            testing.assert_almost_equal(stad.perimeter, pi * width )
            testing.assert_almost_equal(stad.radius, 0.5 * width)
            testing.assert_almost_equal(stad.thickness, 0.0)
            testing.assert_almost_equal(stad.width, width)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            depth = (3.15 - 2.0 * width) / (pi - 2.0)
            stad = Stadium('Lb1: mid-arm', 'depthwidth', depth, width)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            testing.assert_almost_equal(stad.perimeter, pi * width )
            testing.assert_almost_equal(stad.radius, 0.5 * width)
            testing.assert_almost_equal(stad.thickness, 0.0)
            testing.assert_almost_equal(stad.width, width)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertRaises(ValueError, Stadium, 'Lb1: mid-arm',
                    'thicknessradius', -.1, -.5)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertRaises(ValueError, Stadium, 'Lb1: mid-arm',
                    'thicknessradius', 1.0, -.3)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.assertRaises(ValueError, Stadium, 'Lb1: mid-arm',
                    'thicknessradius', -.1, 2)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)

        # Radius cannot be zero.
        self.assertRaises(ValueError, Stadium, 'Lb1: mid-arm',
                'thicknessradius', -.1, 0)
        self.assertRaises(ValueError, Stadium, 'Lb1: mid-arm',
                'thicknessradius', 0, 0)

        sys.stdout = actual_stdout


def test_solid():
    label = 'Test'
    density = 3.0
    height = 5.0
    sol = Solid(label, density, height)
    assert sol.label == label
    testing.assert_almost_equal(sol.density, density)
    testing.assert_almost_equal(sol.height, height)
    position = array([1., 2., 3.])
    # body-three 1-2-3
    angles = array([0.34, 23.6, -0.2])
    c1 = cos(angles[0])
    c2 = cos(angles[1])
    c3 = cos(angles[2])
    s1 = sin(angles[0])
    s2 = sin(angles[1])
    s3 = sin(angles[2])
    # definition of body 1-2-3 rotations from Spacecraft Dynamics, Kane,
    # Likins, Levinson, 1982 page 423 (this is the transpose of what is
    # presented)
    C = matrix([[c2 * c3, s1 * s2 * c3 + s3 * c1, -c1 * s2 * c3 + s3 * s1],
                [-c2 * s3, -s1 * s2 * s3 + c3 * c1, c1 * s2 * s3 + c3 *s1],
                [s2, -s1 * c2, c1 * c2]])

    sol.set_orientation(position, C, True)
    testing.assert_allclose(sol.pos, position)
    testing.assert_allclose(sol._rot_mat, C)
    testing.assert_allclose(sol.end_pos, position + (height * C * array([[0],
        [0], [1]])))
    testing.assert_allclose(sol.inertia, zeros((3, 3)))

    #TODO: complete tests for solid and the remaining classes in solid.py

def test_stadiumsolid_inertial_properties():
    """Checks the inertial property calculations of the StadiumSolid."""

    density = 1.5
    height = 4
    rad0 = 3
    thick0 = 1
    rad1 = 1
    thick1 = 2

    # Create stadiumsolid using the yeadon package.
    stad1 = Stadium('Ls1: umbilicus', 'thicknessradius', thick0, rad0)
    stad2 = Stadium('Lb1: mid-arm', 'thicknessradius', thick1, rad1)
    solid = StadiumSolid('solid', density, stad1, stad2, height)

    # Create stadiumsolid for checking.
    solid_des = StadiumSolidCheck(density, thick0, rad0, thick1, rad1, height)

    testing.assert_almost_equal(solid.mass, solid_des.mass())
    testing.assert_almost_equal(solid.rel_center_of_mass[2],
            solid_des.mass_center())
    testing.assert_almost_equal(solid.rel_inertia[2, 2], solid_des.inertia_zz())
    testing.assert_almost_equal(solid.rel_inertia[1, 1], solid_des.inertia_yy())
    testing.assert_almost_equal(solid.rel_inertia[0, 0], solid_des.inertia_xx())

    # Switch the order of the stadia and ensure everything still works out.
    solid = StadiumSolid('solid', density, stad2, stad1, height)

    # Create stadiumsolid for checking.
    solid_des = StadiumSolidCheck(density, thick1, rad1, thick0, rad0, height)

    testing.assert_almost_equal(solid.mass, solid_des.mass())
    testing.assert_almost_equal(solid.rel_center_of_mass[2],
            solid_des.mass_center())
    testing.assert_almost_equal(solid.rel_inertia[2, 2], solid_des.inertia_zz())
    testing.assert_almost_equal(solid.rel_inertia[1, 1], solid_des.inertia_yy())
    testing.assert_almost_equal(solid.rel_inertia[0, 0], solid_des.inertia_xx())

def test_stadiumsolidcheck_symmetry():
    """Tests the symmetry of the stadiumsolid formulae, as Yeadon presented
    them (not as implemented). That means, if we switch which stadium we call 0
    or 1, we look at if the mass/volume, center of mass, moments of inertia
    change.

    """
    density = 1.5
    height = 4

    # Same top and bottom.
    r = 3; t = 2;
    solid_desA = StadiumSolidCheck(density, t, r, t, r, height)
    testing.assert_almost_equal(solid_desA.mass(),
            density * height * (4 * r * t + pi * r**2))

    # Diff r, same t.
    r0 = 3; t0 = 2; r1 = 2; t1 = 2;
    solid_desA = StadiumSolidCheck(density, t0, r0, t1, r1, height)
    solid_desB = StadiumSolidCheck(density, t1, r1, t0, r0, height)
    testing.assert_almost_equal(solid_desA.mass(), solid_desB.mass())
    testing.assert_almost_equal(solid_desA.mass_center(),
            height - solid_desB.mass_center())
    testing.assert_almost_equal(solid_desA.inertia_zz(), solid_desB.inertia_zz())
    testing.assert_almost_equal(solid_desA.inertia_yy(), solid_desB.inertia_yy())
    testing.assert_almost_equal(solid_desA.inertia_xx(), solid_desB.inertia_xx())

    # Same r, diff t.
    r0 = 3; t0 = 2; r1 = 3; t1 = 1;
    solid_desA = StadiumSolidCheck(density, t0, r0, t1, r1, height)
    solid_desB = StadiumSolidCheck(density, t1, r1, t0, r0, height)
    testing.assert_almost_equal(solid_desA.mass(), solid_desB.mass())
    testing.assert_almost_equal(solid_desA.mass_center(),
            height - solid_desB.mass_center())
    testing.assert_almost_equal(solid_desA.inertia_zz(), solid_desB.inertia_zz())
    testing.assert_almost_equal(solid_desA.inertia_yy(), solid_desB.inertia_yy())
    testing.assert_almost_equal(solid_desA.inertia_xx(), solid_desB.inertia_xx())

    # Diff r, diff t, one is included in the other.
    r0 = 3; t0 = 2; r1 = 2; t1 = 1;
    solid_desA = StadiumSolidCheck(density, t0, r0, t1, r1, height)
    solid_desB = StadiumSolidCheck(density, t1, r1, t0, r0, height)
    testing.assert_almost_equal(solid_desA.mass(), solid_desB.mass())
    testing.assert_almost_equal(solid_desA.mass_center(),
            height - solid_desB.mass_center())
    testing.assert_almost_equal(solid_desA.inertia_zz(), solid_desB.inertia_zz())
    testing.assert_almost_equal(solid_desA.inertia_yy(), solid_desB.inertia_yy())
    testing.assert_almost_equal(solid_desA.inertia_xx(), solid_desB.inertia_xx())

    # Diff r, diff t, overlap.
    r0 = 3; t0 = 1; r1 = 2; t1 = 5;
    solid_desA = StadiumSolidCheck(density, t0, r0, t1, r1, height)
    solid_desB = StadiumSolidCheck(density, t1, r1, t0, r0, height)
    testing.assert_almost_equal(solid_desA.mass(), solid_desB.mass())
    testing.assert_almost_equal(solid_desA.mass_center(),
            height - solid_desB.mass_center())
    testing.assert_almost_equal(solid_desA.inertia_zz(), solid_desB.inertia_zz())
    testing.assert_almost_equal(solid_desA.inertia_yy(), solid_desB.inertia_yy())
    testing.assert_almost_equal(solid_desA.inertia_xx(), solid_desB.inertia_xx())

def test_degenerate_stadiumsolid_symmetry():
    """Tests the validity, and symmetry, of the stadiumsolid formulae, as
    implemented with the t0 == 0 correction. That means, if we switch which
    stadium we call 0 or 1, we look at if the mass/volume, center of mass,
    moments of inertia change.

    """
    density = 1.5
    height = 4
    height_vec = array([[0], [0], [height]])

    # One thickness is 0.
    r0 = 5; t0 = 0; r1 = 2; t1 = 2;
    # For checking against.
    stad1_des = Stadium('Ls1: umbilicus', 'thicknessradius', 0.000000001, r0)

    stad1 = Stadium('Ls1: umbilicus', 'thicknessradius', t0, r0)
    stad2 = Stadium('Lb1: mid-arm', 'thicknessradius', t1, r1)

    solidA = StadiumSolid('solid', density, stad1_des, stad2, height)
    solidA_des= StadiumSolid('solid', density, stad1, stad2, height)
    solidB = StadiumSolid('solid', density, stad2, stad1_des, height)
    solidB_des = StadiumSolid('solid', density, stad2, stad1, height)

    testing.assert_almost_equal(solidB.mass, solidB_des.mass, decimal=4)
    testing.assert_allclose(solidB.rel_center_of_mass,
            solidB_des.rel_center_of_mass)
    testing.assert_almost_equal(solidB.rel_inertia[0,0],
            solidB_des.rel_inertia[0,0], decimal=4)
    testing.assert_almost_equal(solidB.rel_inertia[1,1],
            solidB_des.rel_inertia[1,1], decimal=4)
    testing.assert_almost_equal(solidB.rel_inertia[2,2],
            solidB_des.rel_inertia[2,2], decimal=4)

    testing.assert_almost_equal(solidA.mass, solidA_des.mass)
    testing.assert_allclose(solidA.rel_center_of_mass,
            solidA_des.rel_center_of_mass)
    testing.assert_almost_equal(solidA.rel_inertia[0,0],
            solidA_des.rel_inertia[0,0], decimal=4)
    testing.assert_almost_equal(solidA.rel_inertia[1,1],
            solidA_des.rel_inertia[1,1], decimal=4)
    testing.assert_almost_equal(solidA.rel_inertia[2,2],
            solidA_des.rel_inertia[2,2], decimal=4)

    testing.assert_almost_equal(solidA.mass, solidB.mass)
    testing.assert_allclose(solidA.rel_center_of_mass,
            height_vec - solidB.rel_center_of_mass)
    testing.assert_almost_equal(solidA.rel_inertia[0,0], solidB.rel_inertia[0,0])
    testing.assert_almost_equal(solidA.rel_inertia[1,1], solidB.rel_inertia[1,1])
    testing.assert_almost_equal(solidA.rel_inertia[2,2], solidB.rel_inertia[2,2])

    # Both thicknesses are zero.
    r0 = 3; t0 = 0; r1 = 2; t1 = 0;
    stad1 = Stadium('Ls1: umbilicus', 'thicknessradius', t0, r0)
    stad2 = Stadium('Lb1: mid-arm', 'thicknessradius', t1, r1)

    solidA = StadiumSolid('solid', density, stad1, stad2, height)
    solidB = StadiumSolid('solid', density, stad2, stad1, height)

    testing.assert_almost_equal(solidA.mass, solidB.mass)
    testing.assert_allclose(solidA.rel_center_of_mass,
            height_vec - solidB.rel_center_of_mass)
    testing.assert_almost_equal(solidA.rel_inertia[0,0], solidB.rel_inertia[0,0])
    testing.assert_almost_equal(solidA.rel_inertia[1,1], solidB.rel_inertia[1,1])
    testing.assert_almost_equal(solidA.rel_inertia[2,2], solidB.rel_inertia[2,2])

    # A third case for when both t0 and t0 are zero.
    # TODO

def test_stadiumsolidcheck_against_truncated_cone():
    """Tests the StadiumSolidCheck formulae above against truncated cone
    formulae for degenerate stadia; using a thin trapezium."""

    def truncated_cone_mass(density, radius0, radius1, height):
        return density / 3.0 * pi * height * (radius0**2 + radius1**2 +
                radius0 * radius1)

    density = 1.5
    height = 4
    rad0 = 3
    thick0 = 0.0000001
    rad1 = 4
    thick1 = 0.0000001

    # Create stadiumsolid for checking.
    solid_des = StadiumSolidCheck(density, thick0, rad0, thick1, rad1, height)

    testing.assert_almost_equal(solid_des.mass(),
            truncated_cone_mass(density, rad0, rad1, height), decimal=4)

    # Now only one level is a circle, so we'll add in an extruded triangle in the
    # middle to find its volume on our own (to check). Radii must be the same
    # for this to work out in the simple case.
    thick0 = 1
    rad1 = 3
    solid_des2 = StadiumSolidCheck(density, thick0, rad0, thick1, rad1, height)

    testing.assert_almost_equal(solid_des2.mass(),
            truncated_cone_mass(density, rad0, rad1, height) +
            density * (2 * thick0 * height * 0.5 * (rad0 * 2)), decimal=4)

def test_rotate_inertia():
    """Are we obtaining the global inertia properly?"""

    density = 1.5
    height = 4
    height_vec = array([[0], [0], [height]])

    # One thickness is 0.
    r0 = 5; t0 = 0; r1 = 2; t1 = 2;

    stad1 = Stadium('Ls1: umbilicus', 'thicknessradius', t0, r0)
    stad2 = Stadium('Lb1: mid-arm', 'thicknessradius', t1, r1)

    solidA = StadiumSolid('solid', density, stad1, stad2, height)

    # This inertia matrix describes two 1kg point masses at (0, 2, 1) and
    # (0, -2, -1) in the global reference frame, A.
    solidA._rel_inertia = mat([[10.0, 0.0, 0.0],
                             [0.0, 2.0, -4.0],
                             [0.0, -4.0, 8.0]])

    # If we want the inertia about a new reference frame, B, such that the
    # two masses lie on the yb axis we can rotate about xa through the angle
    # arctan(1/2). Note that this function returns R from va = R * vb.
    solidA._rot_mat = inertia.rotate_space_123((arctan(1.0 / 2.0), 0.0, 0.0))

    solidA.calc_properties()

    I_b = solidA.inertia

    expected_I_b = mat([[10.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0],
                        [0.0, 0.0, 10.0]])

    testing.assert_allclose(I_b, expected_I_b, atol=1e-16)



"""Solid objects are used by the segment module. A solid has a position, and
orientation (defined by a rotation matrix). This module also contains the
class definition for stadium objects, which are used to construct
StadiumSolid solids. The Solid class has two subclasses: the StadiumSolid and
Semiellipsoid classes.

"""
# Use Python3 integer division rules.
from __future__ import division
import warnings

import numpy as np

from . import inertia
from .utils import printoptions

class Stadium(object):
    """Stadium, the 2D shape.

    """
    validStadiaLabels = {
        'Ls0': 'hip joint centre',
        'Ls1': 'umbilicus',
        'Ls2': 'lowest front rib',
        'Ls3': 'nipple',
        'Ls4': 'shoulder joint centre',
        'Ls5': 'acromion',
        'Ls6': 'beneath nose',
        'Ls7': 'above ear',
        'La0': 'shoulder joint centre',
        'La1': 'mid-arm',
        'La2': 'elbow joint centre',
        'La3': 'maximum forearm perimeter',
        'La4': 'wrist joint centre',
        'La5': 'base of thumb',
        'La6': 'knuckles',
        'La7': 'fingernails',
        'Lb0': 'shoulder joint centre',
        'Lb1': 'mid-arm',
        'Lb2': 'elbow joint centre',
        'Lb3': 'maximum forearm perimeter',
        'Lb4': 'wrist joint centre',
        'Lb5': 'base of thumb',
        'Lb6': 'knuckles',
        'Lb7': 'fingernails',
        'Lj0': 'hip joint centre',
        'Lj1': 'crotch',
        'Lj2': 'mid-thigh',
        'Lj3': 'knee joint centre',
        'Lj4': 'maximum calf perimeter',
        'Lj5': 'ankle joint centre',
        'Lj6': 'heel',
        'Lj7': 'arch',
        'Lj8': 'ball',
        'Lj9': 'toe nails',
        'Lk0': 'hip joint centre',
        'Lk1': 'crotch',
        'Lk2': 'mid-thigh',
        'Lk3': 'knee joint centre',
        'Lk4': 'maximum calf perimeter',
        'Lk5': 'ankle joint centre',
        'Lk6': 'heel',
        'Lk7': 'arch',
        'Lk8': 'ball',
        'Lk9': 'toe nails'}

    def __init__(self, label, inID, in1, in2=None, alignment='ML'):
        """Defines a 2D stadium shape and checks inputs for errors. A stadium,
        described in Yeadon 1990-ii, is defined by two parameters. Stadia can
        depracate to circles if their "thickness" is 0.

        Parameters
        ----------
        label : str
            Name of the stadium level, according to Yeadon 1990-ii.
        inID : str
            Identifies the type of information for the next two inputs.
            'perimwidth' for perimeter and width input, 'depthwidth' for
            depth and width input, 'perimeter' or 'radius' for a circle,
            'thicknessradius' for thickness and radius input.
        in1 : float
            Either perimeter, depth, or thickness, as determined by inID
        in2 : float
            Either width, or radius, as determined by inID
        alignment = 'ML' : str
            Identifies the long direction of the stadium. 'ML' stands for
            medio-lateral. Aleternatively, 'AP' (anteroposterior) can be
            supplied. The only 'AP' stadiums should be at the heels.

        """
        if label == 'Ls5: acromion/bottom of neck':
            self.label = label
        elif label in [lab + ': ' + desc for lab, desc in
                self.validStadiaLabels.items()]:
            self.label = label
        else:
            raise ValueError("'{}' is not a valid label.".format(label))

        if inID == 'perimwidth':
            self.perimeter = in1
            self.width = in2
            self.thickness = ((np.pi * self.width - self.perimeter) /
                          (2.0 * np.pi - 4.0))
            self.radius = ((self.perimeter - 2.0 * self.width)  /
                           (2.0 * np.pi - 4.0))
        elif inID == 'depthwidth':
            self.width = in2
            self.perimeter = 2.0 * in2 + (np.pi - 2.0) * in1
            self.thickness = ((np.pi * self.width - self.perimeter) /
                          (2.0 * np.pi - 4.0))
            self.radius = (self.perimeter - 2.0 * self.width) / (2.0 * np.pi -
                    4.0)
        elif inID == 'perimeter':
            self._set_as_circle(in1 / (2.0 * np.pi))
        elif inID == 'radius':
            self._set_as_circle(in1)
        elif inID == 'thicknessradius':
            self.thickness = in1
            self.radius = in2
            self.perimeter = 4.0 * self.thickness + 2.0 * np.pi * self.radius
            self.width = 2.0 * self.thickness + 2.0 * self.radius
        else:
            raise ValueError("Stadium " + self.label +
                " not defined properly, " + inID + " is not valid. You must " +
                "use inID= perimwidth, depthwidth, perimeter, or radius.")
        if self.radius == 0:
            raise ValueError("Radius of stadium '%' is zero." % self.radius)
        if self.radius < 0 or self.thickness < 0:
            warnings.warn("Stadium '{}' is defined "
                "incorrectly, r must be positive and t must be nonnegative. "
                "r = {} and t = {} . This means that 2 < perimeter/width < pi. "
                "Currently, this ratio is {}.\n".format(self.label,
                    self.radius, self.thickness, self.perimeter / self.width))
            if inID == 'perimwidth':
                self._set_as_circle(in1 / (2.0 * np.pi))
                print("Fix: stadium set as circle with perimeter as given.")
            elif inID == 'depthwidth':
                self._set_as_circle(0.5 * in2)
                print("Fix: stadium set as circle with diameter of given "
                        "width.")
            else:
                raise ValueError("Negative radius/thickness cannot be "
                        "corrected.")
        if alignment != 'AP' and alignment != 'ML':
            raise ValueError("Error: stadium " + self.label +
                " alignment is not valid, must be either AP or ML")
        else:
            self.alignment = alignment

    def _set_as_circle(self, radius):
        """Sets radius, perimeter, thickness, and width if thickness is 0."""
        self.radius = radius
        self.perimeter = 2.0 * np.pi * self.radius
        self.thickness = 0.0
        self.width = self.perimeter / np.pi

class Solid(object):
    """Solid. Has two subclasses, stadiumsolid and semiellipsoid. This base
    class manages setting orientation, and calculating properties.

    """
    # Transparency for plotting.
    alpha = .5

    @property
    def mass(self):
        """Mass of the solid, a float in units of kg."""
        return self._mass

    @property
    def center_of_mass(self):
        """Center of mass of the solid, a np.ndarray of shape (3,1), in
        units of m, expressed in the global frame, from the bottom center of
        the pelvis (Ls0)."""
        return self._center_of_mass

    @property
    def inertia(self):
        """Inertia matrix of the solid, a np.matrix of shape (3,3), in units
        of kg-m^2, about the center of mass of the human, expressed in the
        global frame.
        """
        return self._inertia

    @property
    def rel_center_of_mass(self):
        """Center of mass of the solid, a np.ndarray of shape (3,1), in
        units of m, expressed in the frame of the solid, from the origin of
        the solid."""
        return self._rel_center_of_mass

    @property
    def rel_inertia(self):
        """Inertia matrix of the solid, a np.matrix of shape (3,3), in units
        of kg-m^2, about the center of mass of the solid, expressed in the
        frame of the solid."""
        return self._rel_inertia

    @property
    def pos(self):
        """Position of the origin of the solid, which is the center of the
        surface closest to the pelvis, a np.ndarray of shape (3,1), in units
        of m, expressed in the global frame, from the bottom center of the
        pelvis (Ls0)."""
        return self._pos

    @property
    def end_pos(self):
        """Position of the point on the solid farthest from the origin along
        the longitudinal axis of the segment, a np.ndarray of shape (3,1),
        in units of m, expressed in the global frame, from the bottom center
        of the pelvis (Ls0)."""
        return self._end_pos

    def __init__(self, label, density, height):
        """Defines a solid. This is a base class. Sets the alpha value to
        be used for drawing with MayaVi.

        Parameters
        ----------
        label : str
            Name of the solid
        density : float
            In units (kg/m^3), used to calculate the solid's mass
        height : float
            Distance from bottom to top of the solid

        """
        #TODO: Check for valid labels
        self.label = label
        #TODO: Check that these two are floats
        self.density = density
        self.height = height
        self._rel_inertia = np.zeros((3, 3)) # this gets set in subclasses
        self._mass = 0.0
        self._rel_center_of_mass = np.array([[0.0], [0.0], [0.0]])

    def set_orientation(self, proximal_pos, rot_mat, build_toward_positive_z):
        """Sets the position, rotation matrix of the solid, and calculates
        the "absolute" properties (center of mass, and inertia tensor) of the
        solid.

        Parameters
        ----------
        proximal_pos : np.array (3,1)
            Position of center of proximal end of solid in the absolute fixed
            coordinates of the human.
        rot_mat : np.matrix (3,3)
            Orientation of solid, with respect to the fixed coordinate system.
        build_toward_positive_z : bool, optional
            The order of the solids in the parent segment matters. By default
            they are stacked on top of each other in the segment's local +z
            direction. If this is set to False, then they are stacked in the
            local -z direction. This is done so that, for example, in the default
            configuration, the arms are directed down.

        """
        self._rot_mat = rot_mat
        if build_toward_positive_z:
            self._pos = proximal_pos
            self._end_pos = self.pos + (self.height * self._rot_mat *
                                      np.array([[0], [0], [1]]))
        else:
            self._end_pos = proximal_pos
            self._pos = self._end_pos - (self.height * self._rot_mat *
                    np.array([[0], [0], [1]]))
        self.calc_properties()

    def calc_properties(self):
        """Sets the center of mass and inertia of the solid, both with respect
        to the fixed human frame.

        """
        try:
            try:
                # Here is v_a = R * v_b, where A is global frame and B is
                # rotated frame relative to A.
                self._center_of_mass = (self.pos + self._rot_mat *
                        self.rel_center_of_mass)
            except AttributeError as err:
                message = str(err) + \
                    '. You must set the orientation before attempting ' + \
                    'to calculate the properties.'
                raise AttributeError(message)
        except AttributeError as e:
            print(e)

        self._inertia = inertia.rotate_inertia(self._rot_mat, self.rel_inertia)

    def print_properties(self, precision=5, suppress=True):
        """Prints mass, center of mass (in solid and global frames), and
        inertia (in solid and global frames).

        The solid's origin is at the bottom center of the proximal stadium
        (or stadium closest to the pelvis, Ls0).

        Parameters
        ----------
        precision : integer, default=5
            The precision for floating point representation.
        suppress : boolean, default=True
            Print very small values as 0 instead of scientific notation.

        Notes
        -----
        See numpy.set_printoptions for more details on the optional
        arguments.

        """
        # self.COM, etc. needs to be defined first.
        if not hasattr(self, 'center_of_mass') or not hasattr(self, 'inertia'):
            self.calc_properties()

        template = \
"""\
{label} properties:

Mass (kg):

{mass:1.{precision}f}

COM in solid's frame from solid's origin (m):

{rel_center_of_mass}

COM in global frame from bottom center of pelvis (Ls0) (m):

{center_of_mass}

Inertia tensor in solid's frame about solid's COM (kg-m^2):

{rel_inertia}

Inertia tensor in global frame about solid's COM (kg-m^2):

{inertia}
"""

        with printoptions(precision=precision, suppress=suppress):
            print(template.format(label=self.label,
                                  mass=self.mass,
                                  precision=precision,
                                  rel_center_of_mass=self.rel_center_of_mass,
                                  center_of_mass=self.center_of_mass,
                                  rel_inertia=self.rel_inertia,
                                  inertia=self.inertia))

    def draw_mayavi(self, mlabobj, col):
        raise NotImplementedError()


class StadiumSolid(Solid):
    """Stadium solid. Derived from the solid class.

    """
    def __init__(self, label, density, stadium0, stadium1, height):
        """Defines a stadium solid object. Creates its base object, and
        calculates relative/local inertia properties.

        Parameters
        ----------
        label : str
            Name of the solid.
        density : float
            Density of the solid (kg/m^3).
        stadium0 : :py:class:`Stadium`
            Lower stadium of the stadium solid.
        stadium1 : :py:class:`Stadium`
            Upper stadium of the stadium solid.
        height : float
            Distance between the lower and upper stadia.

        """
        super(StadiumSolid, self).__init__(label, density, height)
        self.stads = [stadium0, stadium1]
        self.alignment = 'ML'
        # if either stadium is oriented anteroposteriorly.
        # inertia must be rotated, and the plots must be modified
        if (self.stads[0].alignment == 'AP' or
            self.stads[1].alignment == 'AP'):
            self.alignment = 'AP'
        if stadium0.thickness == 0:
            self.degenerate_by_t0 = True
        else:
            self.degenerate_by_t0 = False
        self.calc_rel_properties()
        self._orig_mesh_points = list()
        self._orig_mesh_points.append(self._make_mesh(0))
        self._orig_mesh_points.append(self._make_mesh(1))

    def calc_rel_properties(self):
        """Calculates mass, relative center of mass, and relative/local
        inertia, according to formulae in Appendix B of Yeadon 1990-ii. If the
        stadium solid is arranged anteroposteriorly, the inertia is rotated
        by pi/2 about the z axis.

        """
        # There are two cases of stadium solid degeneracy to consider:
        # t0 = 0, and t0 = t1 = 0. The degeneracy arises when b has a
        # denominator of 0. The case that t1 = 0 is not an issue, then.
        # The way the case of t0 = 0 is handled is by switching the two stadia.
        # Note that thi affects how the relative center of mass is set, but
        # does not affect the mass or moments of inertia calculations.
        # The case in which t0 = t1 = 0, we set b to 1. That is because t = t0
        # (1 + bz) is going to be zero anyway, since t0 = 0.
        D = self.density
        h = self.height
        if self.degenerate_by_t0:
            # Swap the stadia.
            r0 = self.stads[1].radius
            t0 = self.stads[1].thickness
            r1 = self.stads[0].radius
            t1 = self.stads[0].thickness
        else:
            r0 = self.stads[0].radius
            t0 = self.stads[0].thickness
            r1 = self.stads[1].radius
            t1 = self.stads[1].thickness
        a = (r1 - r0) / r0
        if t0 == 0:
            # Truncated cone, since both thicknesses are zero.
            # b can be anything, because t = t0(1 + bz) = (0)(1 + bz) = 0.
            b = 1
        else:
            b = (t1 - t0) / t0
        self._mass = D * h * r0 * (4.0 * t0 * self._F1(a,b) +
                                  np.pi * r0 * self._F1(a,a))
        zcom = D * (h**2.0) * (4.0 * r0 * t0 * self._F2(a,b) +
                               np.pi * (r0**2.0) * self._F2(a,a)) / self.mass
        if self.degenerate_by_t0 and t0 != 0:
            # We swapped the stadia, and it's not a truncated cone.
            # Must define this intermediate because zcom above is still what
            # must be used for the parallel axis theorem below.
            adjusted_zcom = h - zcom
        else:
            adjusted_zcom = zcom
        self._rel_center_of_mass = np.array([[0.0],[0.0],[adjusted_zcom]])

        # moments of inertia
        Izcom = D * h * (4.0 * r0 * (t0**3.0) * self._F4(a,b) / 3.0 +
                         np.pi * (r0**2.0) * (t0**2.0) * self._F5(a,b) +
                         4.0 * (r0**3.0) * t0 * self._F4(b,a) +
                         np.pi * (r0**4.0) * self._F4(a,a) * 0.5 )
        # CAUGHT AN (minor) ERROR IN YEADON'S PAPER HERE. The Dh^3 in the
        # formula below is missing from the second formula for Iy^0 on page 73
        # of Yeadon1990-ii.
        Iy = (D * h * (4.0 * r0 * (t0**3.0) * self._F4(a,b) / 3.0 +
                       np.pi * (r0**2.0) * (t0**2.0) * self._F5(a,b) +
                       8.0 * (r0**3.0) * t0*self._F4(b,a) / 3.0 +
                      np.pi * (r0**4.0) * self._F4(a,a) * 0.25) +
              D * (h**3.0) * (4.0 * r0 * t0 * self._F3(a,b) +
                              np.pi * (r0**2.0) * self._F3(a,a)))
        Iycom = Iy - self.mass * (zcom**2.0)
        Ix = (D * h * (4.0 * r0 * (t0**3.0) * self._F4(a,b) / 3.0 +
                       np.pi * (r0**4.0) * self._F4(a,a) * 0.25) +
              D * (h**3.0) * (4.0 * r0 * t0 * self._F3(a,b) +
                              np.pi * (r0**2.0) * self._F3(a,a)))
        Ixcom = Ix - self.mass*(zcom**2.0)
        self._rel_inertia = np.mat([[Ixcom,0.0,0.0],
                                  [0.0,Iycom,0.0],
                                  [0.0,0.0,Izcom]])
        if self.alignment == 'AP':
            # rearrange to anterorposterior orientation
            self._rel_inertia = inertia.rotate_inertia(
                    inertia.rotate_space_123([0, 0, np.pi/2]), self.rel_inertia)

    def draw_mayavi(self, mlabobj, col):
        """Draws the initial stadium in 3D using MayaVi.

        Parameters
        ----------
        mlabobj : mayavi.soemthing
            The MayaVi object we can draw on.
        col : tuple (3,)
            Color as an rgb tuple, with values between 0 and 1.

        """
        self._generate_mesh()
        self._mesh = mlabobj.mesh(self._mesh_points['x'], self._mesh_points['y'],
                self._mesh_points['z'], color=col, opacity=Solid.alpha)

    def _update_mayavi(self):
        """Updates the mesh in MayaVi."""
        self._generate_mesh()
        self._mesh.mlab_source.set(x=self._mesh_points['x'],
                y=self._mesh_points['y'], z=self._mesh_points['z'])

    def _generate_mesh(self):
        """Generates grid points for a MayaVi mesh."""
        X0, Y0, Z0 = self._make_pos(0)
        X1, Y1, Z1 = self._make_pos(1)
        Xpts = np.array(np.concatenate( (X0, X1), axis=0))
        Ypts = np.array(np.concatenate( (Y0, Y1), axis=0))
        Zpts = np.array(np.concatenate( (Z0, Z1), axis=0))
        self._mesh_points = {'x': Xpts, 'y': Ypts, 'z': Zpts}

    def _make_mesh(self, i):
        """Generates the un-rotated coordinates of the solid. These values are
        saved at instantiation.

        Parameters
        ----------
        i : int
            Identifies which stadium to generate the mesh points for (the top
            or bottom).

        """
        theta = [np.linspace(0.0,np.pi/2,5)]
        x = self.stads[i].thickness + self.stads[i].radius * np.cos(theta);
        y = self.stads[i].radius * np.sin(theta);
        if self.alignment == 'AP':
            temp = x
            x = y
            y = temp
            del temp
        xrev = x[:, ::-1]
        yrev = y[:, ::-1]
        X = np.concatenate( (x, -xrev, -x, xrev), axis=1)
        Y = np.concatenate( (y, yrev, -y, -yrev), axis=1)
        Z = i*self.height*np.ones((1,20))
        POSES = np.concatenate( (X, Y, Z), axis=0)
        return POSES

    def _make_pos(self, i):
        """Generates coordinates to be used for 3D visualization purposes.

        """
        rotated_points = self._rot_mat * self._orig_mesh_points[i]
        X, Y, Z = np.vsplit(rotated_points, 3)
        X = X + self.pos[0]
        Y = Y + self.pos[1]
        Z = Z + self.pos[2]
        return X, Y, Z

    @staticmethod
    def _F1(a, b):
        """Integration term. See Yeadon 1990-ii Appendix 2."""
        return 1.0 + (a + b) * 0.5 + a * b / 3.0

    @staticmethod
    def _F2(a, b):
        """Integration term. See Yeadon 1990-ii Appendix 2."""
        return 0.5 + (a + b) / 3.0 + a * b * 0.25

    @staticmethod
    def _F3(a, b):
        """Integration term. See Yeadon 1990-ii Appendix 2."""
        return 1.0/3.0 + (a + b) / 4.0 + a * b *0.2

    @staticmethod
    def _F4(a, b):
        """Integration term. See Yeadon 1990-ii Appendix 2."""
        return (1.0 + (a + 3.0 * b) * 0.5 + (a * b + b**2.0) +
                      (3.0 * a * b**2.0 + b**3.0) * 0.25 + a * (b**3.0) * 0.2)

    @staticmethod
    def _F5(a, b):
        """Integration term. See Yeadon 1990-ii Appendix 2."""
        return (1.0 + (a + b) + (a**2.0 + 4.0 * a * b + b**2.0) / 3.0 +
                       a * b * (a + b) * 0.5 + (a**2.0) * (b**2.0) * 0.2)

class Semiellipsoid(Solid):
    """Semiellipsoid."""

    n_mesh_points = 30
    def __init__(self,label,density,baseperim,height):
        """Defines a semiellipsoid (solid) object. Creates its base object, and
        calculates relative/local inertia properties. The base is circular (its
        height axis is pointed upwards), so only 2 parameters are needed to
        define the semiellipsoid.

        Parameters
        ----------
        label : str
            Name of the solid.
        density : float
            Density of the solid (kg/m^3).
        baseperimeter : float
            The base is circular.
        height : float
            The remaining minor axis.

        """
        super(Semiellipsoid, self).__init__(label, density, height)
        self.baseperimeter = baseperim
        self.radius = self.baseperimeter/(2.0*np.pi)
        self.calc_rel_properties()
        self._mesh_x, self._mesh_y, self._mesh_z = self._make_mesh()

    def calc_rel_properties(self):
        """Calculates mass, relative center of mass, and relative/local
        inertia, according to somewhat commonly availble formulae.

        """
        D = self.density
        r = self.radius
        h = self.height
        self._mass = D * 2.0/3.0 * np.pi * (r**2) * h
        self._rel_center_of_mass = np.array([[0.0],[0.0],[3.0/8.0 * h]])
        Izcom = D * 4.0/15.0 * np.pi * (r**4.0) * h
        Iycom = D * np.pi * (2.0/15.0 * (r**2.0) * h * (r**2.0 + h**2.0) -
            3.0/32.0 * (r**2.0) * (h**3.0))
        Ixcom = Iycom
        self._rel_inertia = np.mat([[Ixcom,0.0,0.0],
                                  [0.0,Iycom,0.0],
                                  [0.0,0.0,Izcom]])

    def draw_mayavi(self, mlabobj, col):
        """Draws the semiellipsoid in 3D using MayaVi.

        Parameters
        ----------
        mlabobj :
            The MayaVi object we can draw on.
        col : tuple (3,)
            Color as an rgb tuple, with values between 0 and 1.

        """
        self._generate_mesh()
        self._mesh = mlabobj.mesh(*self._mesh_points, color=col,
                opacity=Solid.alpha)

    def _update_mayavi(self):
        """Updates the mesh in MayaVi."""
        self._generate_mesh()
        self._mesh.mlab_source.set(x=self._mesh_points[0],
                y=self._mesh_points[1], z=self._mesh_points[2])

    def _generate_mesh(self):
        """Generates a mesh for MayaVi."""
        self._mesh_points = self._make_pos()

    def _make_mesh(self):
        """Generates the un-rotated coordinates of the solid. These values are
        saved at instantiation.

        """
        u = np.linspace(0, 2.0 * np.pi, self.n_mesh_points)
        v = np.linspace(0, np.pi / 2.0, self.n_mesh_points)
        x = self.radius * np.outer(np.cos(u), np.sin(v))
        y = self.radius * np.outer(np.sin(u), np.sin(v))
        z = self.height * np.outer(np.ones(np.size(u)), np.cos(v))
        return x, y, z

    def _make_pos(self):
        """Generates coordinates to be used for 3D visualization purposes,
        given the position and orientation of the solid.

        """
        x = np.zeros(self._mesh_x.shape)
        y = np.zeros(self._mesh_y.shape)
        z = np.zeros(self._mesh_z.shape)
        for i in np.arange(self.n_mesh_points):
            for j in np.arange(self.n_mesh_points):
                POS = np.array([
                    [self._mesh_x[i,j]],
                    [self._mesh_y[i,j]],
                    [self._mesh_z[i,j]]])
                POS = self._rot_mat * POS
                x[i,j] = POS[0,0]
                y[i,j] = POS[1,0]
                z[i,j] = POS[2,0]
        x = self.pos[0,0] + x
        y = self.pos[1,0] + y
        z = self.pos[2,0] + z
        return x, y, z



"""Segment objects are used by the human module. A segment has a position, and
an orientation. All constituent solids of a segment have the same orientation.
That is to say that the base of the segment is at a joint in the human. The
user does not interact with this module.

"""
# Use Python3 integer division rules.
from __future__ import division

# external imports
import numpy as np

# local imports
from . import inertia
from .utils import printoptions


class Segment(object):

    @property
    def mass(self):
        """Mass of the segment, in units of kg."""
        return self._mass

    @property
    def center_of_mass(self):
        """Center of mass of the segment, a np.ndarray, in units of m,
        expressed in the global frame, from the bottom center of the pelvis
        (Ls0)."""
        return self._center_of_mass

    @property
    def inertia(self):
        """Inertia matrix of the segment, a np.matrix, in units of kg-m^2,
        about the center of mass of the human, expressed in the global
        frame."""
        return self._inertia

    @property
    def rel_center_of_mass(self):
        """Center of mass of the segment, a np.ndarray, in units of m,
        expressed in the frame of the segment, from the origin of the
        segment."""
        return self._rel_center_of_mass

    @property
    def rel_inertia(self):
        """Inertia matrix/dyadic of the segment, a np.matrix, in units of
        kg-m^2, about the center of mass of the segment, expressed in the frame
        of the segment."""
        return self._rel_inertia

    @property
    def pos(self):
        """Position of the origin of the segment, a np.ndarray, in units of m,
        expressed in the global frame, from the bottom center of the pelvis
        (Ls0)."""
        return self._pos

    @property
    def end_pos(self):
        """Position of the center of the last (farthest from pelvis) stadium in
        this segment, a np.ndarray, in units of m, expressed in the global
        frame, from the bottom center of the pelvis (Ls0)."""
        return self._end_pos

    @property
    def rot_mat(self):
        """Rotation matrix specifying the orientation of this segment relative
        to the orientation of the global frame, a np.matrix, unitless.
        Multiplying a vector expressed in this segment's frame with this
        rotation matrix on the left gives that same vector, but expressed in
        the global frame."""
        return self._rot_mat

    def __init__(self, label, pos, rot_mat, solids, color,
                 build_toward_positive_z=True):
        """Initializes a segment object. Stores inputs as instance variables,
        calculates the orientation of the segment's child solids, and
        calculates the "relative" inertia parameters (mass, center of mass
        and inertia) of the segment.

        Parameters
        ----------
        label : str
            The ID and name of the segment.
        pos : numpy.array, shape(3,1)
            The vector position of the segment's base,
            with respect to the global frame.
        rot_mat : numpy.matrix, shape(3,3)
            The orientation of the segment is given by a rotation matrix that
            specifies the orientation of the segment with respect to the fixed
            human frame. That is, rot_mat is (^N R ^A), where N is the fixed
            human frame, and A is the frame fixed to this segment. If v is a
            vector and (^A v) is its representation in A, then (^N R ^A * ^A v)
            = (^N v) is its representation in N.
        solids : list of solid objects
            The solid objects that compose the segment
        color : tuple (3,)
            Color with which to plot this segment in the plotting functions.
            RGB tuple with float values between 0 and 1.
        build_toward_positive_z : bool, optional
            The order of the solids matters. By default they are stacked on top
            of each other in the segment's local +z direction. If this is set to
            False, then they are stacked in the local -z direction. This is
            done so that, for example, in the default configuration, the arms
            are directed down.

        """
        self.label = label
        if pos.shape != (3, 1):
            raise ValueError("Position must be 3-D.")
        self._pos = pos
        self._rot_mat = np.asmatrix(rot_mat)
        self.solids = solids
        self.nSolids = len(self.solids)
        self.color = color
        self._build_toward_positive_z = build_toward_positive_z
        # must set the position of constituent solids before being able to
        # calculate relative/local properties, or set end_pos/length.
        self._set_orientations()
        if self._build_toward_positive_z:
            self._end_pos = self.solids[-1].end_pos
        else:
            self._end_pos = self.solids[-1].pos
        self.length = np.linalg.norm(self._end_pos - self.pos)
        self.calc_rel_properties()

    def _set_orientations(self):
        """Sets the position (self.pos) and rotation matrix (self.rot_mat)
        for all solids in the segment by calling each constituent
        solid's set_orientation method. The position of the i-th solid,
        expressed in the global frame, is given by the sum
        of the segment's base position and the directed height of all the
        solids of the segment up to the i-th solid.

        """
        # pos and rot_mat for first solid
        self.solids[0].set_orientation(self.pos, self.rot_mat,
                self._build_toward_positive_z)
        # pos and rot_mat for remaining solids
        for i in np.arange(self.nSolids):
            if i != 0:
                if self._build_toward_positive_z:
                    pos = self.solids[i-1].end_pos
                else:
                    pos = self.solids[i-1].pos
                self.solids[i].set_orientation(pos, self.rot_mat,
                        self._build_toward_positive_z)

    def calc_rel_properties(self):
        """Calculates the mass, relative/local center of mass, and
        relative/local inertia tensor (about the segment's center of mass).
        Also computes the center of mass of each constituent solid with
        respect to the segment's base in the segment's reference frame.

        """
        # mass
        self._mass = 0.0
        for s in self.solids:
            self._mass += s.mass
        # relative position of each solid w.r.t. segment orientation and
        # segment's origin
        solidpos = []
        # center of mass of each solid w.r.t. segment orientation and
        # segment's origin
        solidCOM = []
        z_unit_vector = np.array([[0, 0, 1]]).T
        if self._build_toward_positive_z:
            solidpos.append(np.zeros((3, 1)))
            for i in np.arange(self.nSolids):
                if i != 0:
                    solidpos.append( solidpos[i-1] +
                                          self.solids[i-1].height *
                                          z_unit_vector)
            solidCOM.append(self.solids[0].rel_center_of_mass)
            for i in np.arange(self.nSolids):
                if i != 0:
                    solidCOM.append( solidpos[i] +
                                          self.solids[i].rel_center_of_mass)
        else: # not self._build_toward_positive_z
            # solidpos
            last_pos = np.zeros((3, 1))
            for solid in self.solids:
                solidpos.append(last_pos - solid.height * z_unit_vector)
                last_pos = solidpos[-1]
            # solidCOM
            for i in np.arange(self.nSolids):
                solidCOM.append(solidpos[i] +
                        self.solids[i].rel_center_of_mass)
        # TODO above code could be substantially cleaned up.
        # relative center of mass
        relmoment = np.zeros((3, 1))
        for i in np.arange(self.nSolids):
            relmoment += self.solids[i].mass * solidCOM[i]
        self._rel_center_of_mass = relmoment / self.mass
        # relative Inertia
        self._rel_inertia = np.mat(np.zeros((3, 3)))
        for i in np.arange(self.nSolids):
            dist = solidCOM[i] - self.rel_center_of_mass
            self._rel_inertia += np.mat(inertia.parallel_axis(
                                      self.solids[i].rel_inertia,
                                      self.solids[i].mass,
                                      [dist[0, 0], dist[1, 0], dist[2, 0]]))

    def calc_properties(self):
        """Calculates the segment's center of mass with respect to the bottm
        center of the pelvis (Ls0) and the segment's inertia in the global
        frame but about the segment's center of mass.

        """
        # center of mass
        self._center_of_mass = self.pos + self.rot_mat * self.rel_center_of_mass
        # inertia in frame f w.r.t. segment's COM
        self._inertia = inertia.rotate_inertia(self.rot_mat, self.rel_inertia)

    def __str__(self):
        return(self._properties_string())

    def print_properties(self, precision=5, suppress=True):
        """Prints mass, center of mass (in segment and global frames),
        and inertia (in solid and global frames).

        Parameters
        ----------
        precision : integer, default=5
            The precision for floating point representation.
        suppress : boolean, default=True
            Print very small values as 0 instead of scientific notation.

        Notes
        -----
        See numpy.set_printoptions for more details on the optional
        arguments.

        """
        print(self._properties_string())

    def _properties_string(self, precision=5, suppress=True):
        """Prints mass, center of mass (in segment and global frames),
        and inertia (in solid and global frames).

        Parameters
        ----------
        precision : integer, default=5
            The precision for floating point representation.
        suppress : boolean, default=True
            Print very small values as 0 instead of scientific notation.

        Notes
        -----
        See numpy.set_printoptions for more details on the optional
        arguments.

        """
        # self.COM, etc. needs to be defined first.
        if not hasattr(self, 'center_of_mass') or not hasattr(self, 'inertia'):
            self.calc_properties()

        template = \
"""\
{label} properties:

Mass (kg):

{mass:1.{precision}f}

COM in segment's frame from segment's origin (m):

{rel_center_of_mass}

COM in global frame from bottom center of pelvis (Ls0) (m):

{center_of_mass}

Inertia tensor in segment's frame about segment's COM (kg-m^2):

{rel_inertia}

Inertia tensor in global frame about segment's COM (kg-m^2):

{inertia}
"""

        with printoptions(precision=precision, suppress=suppress):
            return template.format(label=self.label,
                                  mass=self.mass,
                                  precision=precision,
                                  rel_center_of_mass=self.rel_center_of_mass,
                                  center_of_mass=self.center_of_mass,
                                  rel_inertia=self.rel_inertia,
                                  inertia=self.inertia)

    def print_solid_properties(self, precision=5, suppress=True):
        """Calls the print_properties() member method of each of this
        segment's solids. See the solid class's definition of
        print_properties(self) for more detail.

        Parameters
        ----------
        precision : integer, default=5
            The precision for floating point representation.
        suppress : boolean, default=True
            Print very small values as 0 instead of scientific notation.

        Notes
        -----
        See numpy.set_printoptions for more details on the optional
        arguments.

        """
        for s in self.solids:
            s.print_properties(precision=precision, suppress=suppress)

    def draw_mayavi(self, mlabobj):
        """Draws in a MayaVi window all the solids within this segment.  """
        for s in self.solids:
            s.draw_mayavi(mlabobj, self.color)

    def _update_mayavi(self):
        """Updates all of the solids in this segment for MayaVi."""
        for s in self.solids:
            s._update_mayavi()



# Use Python3 integer division rules.
from __future__ import division

import warnings

import numpy as np

from .exceptions import YeadonDeprecationWarning

# Display our warnings to the user.
warnings.simplefilter('always', YeadonDeprecationWarning)


def parallel_axis(Ic, m, d):
    '''Returns the moment of inertia of a body about a different point.

    Parameters
    ----------
    Ic : array_like, shape(3,3)
        The moment of inertia about the center of mass of the body with
        respect to an orthogonal coordinate system.
    m : float
        The mass of the body.
    d : array_like, shape(3,) or shape(3, 1) or shape(1, 3)
        The distances along the x, y, and z axes that locate the new point
        relative to the center of mass of the body.

    Returns
    -------
    I : numpy.matrix, shape(3,3)
        The moment of inertia of a body about a point located by the
        distances in `d`.

    '''

    Ic = np.asmatrix(Ic)
    d = np.asmatrix(d).reshape((3, 1))

    a = d[0, 0]
    b = d[1, 0]
    c = d[2, 0]

    dMat = np.zeros((3, 3), dtype=Ic.dtype)

    dMat[0] = np.array([b ** 2 + c ** 2, -a * b, -a * c])
    dMat[1] = np.array([-a * b, c ** 2 + a ** 2, -b * c])
    dMat[2] = np.array([-a * c, -b * c, a ** 2 + b ** 2])

    return Ic + m * dMat


def rotate_space_123(angles):
    """Returns the direction cosine matrix relating a reference frame B
    rotated relative to reference frame A through the x, y, then z axes of
    reference frame A (spaced fixed rotations).

    Parameters
    ----------
    angles : numpy.array or list or tuple, shape(3,)
        Three angles (in units of radians) that specify the orientation of
        a new reference frame with respect to a fixed reference frame.
        The first angle is a pure rotation about the x-axis, the second about
        the y-axis, and the third about the z-axis. All rotations are with
        respect to the initial fixed frame, and they occur in the order x,
        then y, then z.

    Returns
    -------
    R : numpy.matrix, shape(3,3)
        Three dimensional rotation matrix about three different orthogonal axes.

    Notes
    -----

    R = |c2 * c3    s1 * s2 * c3 - s3 * c1   c1 * s2 * c3 + s3 * s1|
        |c2 * s3    s1 * s2 * s3 + c3 * c1   c1 * s2 * s3 - c3 * s1|
        |-s2        s1 * c2                  c1 * c2               |

    where

    s1, s2, s3 = sine of the first, second and third angles, respectively
    c1, c2, c3 = cosine of the first, second and third angles, respectively

    So the unit vector b1 in the B frame can be expressed in the A frame (unit
    vectors a1, a2, a3) with:

    b1 = c2 * c3 * a1 + c2 * s3 * a2 - s2 * a3

    Thus a vector vb which is expressed in frame B can be expressed in A by
    pre-multiplying by R:

    va = R * vb

    """
    cx = np.cos(angles[0])
    sx = np.sin(angles[0])

    cy = np.cos(angles[1])
    sy = np.sin(angles[1])

    cz = np.cos(angles[2])
    sz = np.sin(angles[2])

    Rz = np.mat([[ cz,-sz,  0],
                 [ sz, cz,  0],
                 [  0,  0,  1]])

    Ry = np.mat([[ cy,  0, sy],
                 [  0,  1,  0],
                 [-sy,  0, cy]])

    Rx = np.mat([[  1,  0,  0],
                 [  0, cx, -sx],
                 [  0, sx,  cx]])

    return Rz * Ry * Rx

def euler_123(angles):
    """
    Returns the direction cosine matrix of rotated frame B with respect to
    frame A as a function of the Euler 123 angles (body fixed rotation).

    Parameters
    ----------
    angles : array_like, shape(3,)
        Three angles (in units of radians) that specify the orientation of a
        new reference frame, B, with respect to a fixed reference frame, A. The
        first angle, phi, is a rotation about the fixed frame's x-axis. The
        second angle, theta, is a rotation about the new y-axis (which is
        realized after the phi rotation). The third angle, psi, is a rotation
        about the new z-axis (which is realized after the theta rotation).
        Thus, all three angles are "relative" rotations with respect to each
        new frame.  Note: if the rotations are viewed as occuring in the
        opposite direction (z, then y, then x), all three rotations are with
        respect to the initial fixed frame rather than "relative".

    Returns
    -------
    R : numpy.matrix, shape(3,3)
        Three dimensional rotation matrix about three different orthogonal axes.

    Notes
    -----

    R = | c2 * c3                  -c2 * s3                   s2     |
        | s1 * s2 * c3 + s3 * c1   -s1 * s2 * s3 + c3 * c1   -s1 * c2|
        |-c1 * s2 * c3 + s3 * s1    c1 * s2 * s3 + c3 * s1    c1 * c2|

    where

    s1, s2, s3 = sine of the first, second, and third angles, respectively
    c1, c2, c3 = cosine of the first, second, and third angles, respectively

    So the unit vector b1 in the B frame can be expressed in the A frame (unit
    vectors a1, a2, a3) with:

    b1 = c2 * c3 * a1 + (s1 * s2 * c3 + s3 * c1) * a2 +
         (-c1 * c2 * c3 + s3 * s1) * a3

    The rotation matrix is defined such that a R times a vector v_b in the
    rotated reference frame equals the same vector, v_a, expressed in the
    unrotated reference frame.

        R * v_b = v_a

    Where v_a is the vector expressed in the original fixed reference frame and
    v_b is the same vector expressed in the rotated reference frame.

    """

    cphi = np.cos(angles[0])
    sphi = np.sin(angles[0])

    cthe = np.cos(angles[1])
    sthe = np.sin(angles[1])

    cpsi = np.cos(angles[2])
    spsi = np.sin(angles[2])

    R1 = np.mat([[     1,     0,     0],
                 [     0,  cphi, -sphi],
                 [     0,  sphi,  cphi]])

    R2 = np.mat([[  cthe,     0,  sthe],
                 [     0,     1,     0],
                 [ -sthe,     0,  cthe]])

    R3 = np.mat([[  cpsi,  -spsi,     0],
                 [  spsi,  cpsi,     0],
                 [     0,     0,     1]])

    return R1 * R2 * R3


def rotate3_inertia(rotation_matrix, inertia):

    __doc__ = rotate_inertia.__doc__

    # TODO : Remove this function in Yeadon 2.0.

    msg = ("rotate3_inertia has been renamed to rotate_inertia, this " +
           "function signature will be removed in Yeadon 2.0.")
    warnings.warn(msg, YeadonDeprecationWarning)

    return rotate_inertia(rotation_matrix, inertia)


def rotate_inertia(rotation_matrix, inertia):
    """Returns an inertia tensor expressed in a reference frame which has
    been rotated with respect to the frame the inertia tensor is currently
    expressed in.

    Parameters
    ----------
    rotation_matrix : numpy.matrix, shape(3,3)
        Three-dimensional rotation/transformation/direction-cosine matrix
        that transforms a vector in the rotated reference frame into one in
        the current reference frame.
    inertia : numpy.matrix, shape(3,3)
        Three-dimensional cartesian tensor describing the inertia of a rigid
        body in a reference frame.

    Returns
    -------
    rotated_inertia : numpy.matrix, shape(3,3)
        The inertia tensor expressed in the rotated reference frame.

    Notes
    -----

    The provided inertia tensor is expressed in a reference frame, A, and
    there is a reference frame, B, which is rotated with respect to A such
    that a vector, v, expressed in B as v_b can be expressed in A as v_a by
    pre-multiplying by the rotation matrix, R:

    v_a = R * v_b

    Angular momentum of a rigid body expressed in A is defined as:

    H_a = I_a * w_a

    where H_a and w_a are the angular momentum and angular rate vectors,
    respectively and I_a is the inertia tensor, all expressed in A.
    Expressing H_a and w_a in B gives:

    R * H_b = I_a * R * w_b

    R * I_b * w_b = I_a * R * w_b

    So,

    R * I_b = I_a * R

    and thus:

    I_b = R^T * I_a * R

    """
    return rotation_matrix.T * inertia * rotation_matrix


def total_com(coordinates, masses):
    """
    Returns the center of mass of a group of objects if the indivdual
    centers of mass and mass is provided.

    coordinates : ndarray, shape(3,n)
        The rows are the x, y and z coordinates, respectively and the columns
        are for each object.
    masses : ndarray, shape(3,)
        An array of the masses of multiple objects, the order should correspond
        to the columns of coordinates.

    Returns
    -------
    mT : float
        Total mass of the objects.
    cT : ndarray, shape(3,)
        The x, y, and z coordinates of the total center of mass.

    """
    products = masses * coordinates
    mT = np.sum(masses)
    cT = np.sum(products, axis=1) / mT
    return mT, cT

def principal_axes(I):
    """
    Returns the principal moments of inertia and the orientation.

    Parameters
    ----------
    I : ndarray, shape(3,3)
        An inertia tensor.

    Returns
    -------
    Ip : ndarray, shape(3,)
        The principal moments of inertia. This is sorted smallest to largest.
    C : ndarray, shape(3,3)
        The rotation matrix.

    """
    Ip, C = np.linalg.eig(I)
    indices = np.argsort(Ip)
    Ip = Ip[indices]
    C = C.T[indices]
    return Ip, C



#!/usr/bin/env python


class YeadonDeprecationWarning(DeprecationWarning):
    """Simple wrapper so that our deprecation warnings are shown to the
    user. By default, DeprecationWarning's are not printed."""
    pass
